__all__ = ('Menu',)

from collections import OrderedDict
from types import MemberDescriptorType

from scarletio import CallableAnalyzer, CancelledError, RichAttributeErrorBaseType, Task, copy_docs

from ....discord.allowed_mentions import AllowedMentionProxy
from ....discord.channel import Channel
from ....discord.client import Client
from ....discord.core import KOKORO
from ....discord.embed import Embed
from ....discord.exceptions import DiscordException, ERROR_CODES
from ....discord.component import Component, ComponentType, create_row
from ....discord.interaction import  InteractionEvent, InteractionType
from ....discord.message import Message

from ..waiters import Timeouter


GUI_STATE_NONE = 0
GUI_STATE_READY = 1
GUI_STATE_EDITING = 2
GUI_STATE_CANCELLING = 3
GUI_STATE_CANCELLED = 4
GUI_STATE_SWITCHING_CONTEXT = 5

INTERACTION_TYPE_APPLICATION_COMMAND = InteractionType.application_command
INTERACTION_TYPE_MESSAGE_COMPONENT = InteractionType.message_component

class ComponentSourceIdentityHasher:
    """
    Hasher for components based on their identity.
    
    Attributes
    ----------
    component : ``Component``
        The stored component by the hasher.
    """
    __slots__ = ('component', )
    
    def __init__(self, component):
        """
        Creates a new ``ComponentSourceIdentityHasher`` with the given parameters.
        
        Parameters
        ----------
        component : ``Component``
            The source component.
        """
        self.component = component
    
    def __hash__(self):
        """Returns the component source identity hasher's hash value."""
        return object.__hash__(self.component)
    
    def __eq__(self, other):
        """Returns whether the two component identity hashers are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.component is other.component



class ComponentDescriptor:
    """
    Descriptor to proxy class attribute component access for each instance.
    
    Attributes
    ----------
    _source_component : ``Component``
        The source component from which the descriptor is created from.
    _sub_component_descriptors : `None`, `list` of ``ComponentDescriptor``
        Sub components applicable to the component.
    _identifier : `int`
        The descriptor's identifier.
    
    Class Attributes
    ----------------
    _identifier_counter : `int`
        Identifier counter for caching.
    """
    __slots__ = ('_source_component', '_sub_component_descriptors', '_identifier',)
    
    _identifier_counter = 0
    
    
    def __new__(cls, source_component, sub_components):
        """
        Creates a new ``ComponentDescriptor`` from the given component.
        
        Parameters
        ----------
        source_component : ``Component``
            The source component to create the descriptor from.
        sub_component_descriptors : `None`, `list` of ``ComponentDescriptor``
            A list of sub components added
        """
        identifier = cls._identifier_counter + 1
        cls._identifier_counter = identifier
        
        self = object.__new__(cls)
        self._source_component = source_component
        self._sub_component_descriptors = sub_components
        self._identifier = identifier
        return self
    
    
    def __get__(self, instance, type_):
        """
        Gets the descriptor itself if called from class or a component proxy.
        
        Parameters
        ----------
        instance : `None`, ``Menu``
            The menu instance.
        type_ : ``MenuType``
            The menu's type.
        
        Returns
        -------
        self / component : ``ComponentDescriptor`` / ``Component``
            Self if called from class, or the a component proxy for the represented component.
        """
        if instance is None:
            return self
        
        component_cache = instance._component_cache
        try:
            component = component_cache[self._identifier]
        except KeyError:
            component = self._create_component(instance)
            component_cache[self._identifier] = component
        
        return component
    
    
    def __set__(self, instance, value):
        """
        Sets a new component overwriting the already existing value.
        
        Parameters
        ----------
        instance : `None`, ``Menu``
            The menu instance.
        value : ``Component``
            The component to set.
        
        Raises
        ------
        TypeError
            - If `value` is not a ``Component`` instance.
        """
        if not isinstance(value, Component):
            raise TypeError(
                f'Attribute can be set only as `{Component.__name__}`, got {value.__class__.__name__}; {value!r}.'
            )
            
        instance._component_cache[self._identifier] = value
    
    
    def _create_component(self, instance):
        """
        Crates the component of the descriptor.
        
        Parameters
        ----------
        instance : ``Menu``
            The menu instance.
        
        Returns
        -------
        component : ``Component``
        """
        sub_component_descriptors = self._sub_component_descriptors
        source_component = self._source_component
        if (sub_component_descriptors is None):
            return source_component.copy()
        
        return source_component.copy_with(
            components = [
                sub_component_descriptor.__get__(instance, type(instance))
                for sub_component_descriptor in sub_component_descriptors
            ],
        )
    
    def __repr__(self):
        """Returns the component descriptor's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' identifier=', repr(self._identifier),
            ', source_component=', repr(self._source_component),
        ]
        
        sub_component_descriptors = self._sub_component_descriptors
        if (sub_component_descriptors is not None):
            repr_parts.append(' sub_component_descriptors=')
            repr_parts.append(repr(sub_component_descriptors))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


def validate_check(check):
    """
    Validates the given check.
    
    Parameters
    ----------
    check : `None` of `callable`
        The check to validate.
    
    Raises
    ------
    TypeError
        If `check` is not `None` neither a non-async function accepting 1 parameter.
    """
    if check is None:
        return
    
    analyzer = CallableAnalyzer(check)
    if analyzer.is_async():
        raise TypeError(
            f'`check` should have NOT be an `async` function, got {check!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(
            f'`check` should accept `2` parameters, meanwhile the given one expects at '
            f'least `{min_!r}`, got {check!r}.'
        )
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_positional_parameters():
                raise TypeError(
                    f'`check` should accept `2` parameters, meanwhile the given one expects '
                    f'up to `{max_!r}`, got {check!r}.'
                )


def validate_invoke(invoke):
    """
    Validates the given invoker.
    
    Parameters
    ----------
    invoke : `None`, `async-callable`
        The invoker to validate.
    
    Raises
    ------
    TypeError
        If `invoke` is not `None` nor `async-callable` accepting `2` parameter.
    """
    if invoke is None:
        return
    
    analyzer = CallableAnalyzer(invoke)
    if not analyzer.is_async():
        raise TypeError(
            f'`invoke` should be an `async` function, got {invoke!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    
    if min_ > 2:
        raise TypeError(
            f'`invoke` should accept `2` parameters, meanwhile the given one expects at '
            f'least `{min_!r}`, got {invoke!r}.'
        )
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_positional_parameters():
                raise TypeError(
                    f'`invoke` should accept `2` parameters, meanwhile the given one expects '
                    f'up to `{max_!r}`, got {invoke!r}.'
                )


def validate_initial_invoke(initial_invoke):
    """
    Validates the given default content getter.
    
    Parameters
    ----------
    initial_invoke : `None`, `async-callable`
        The default content getter to validate.
    
    Raises
    ------
    TypeError
        If `initial_invoke` is not `None` nor `async-callable` accepting `1` parameters.
    """
    if initial_invoke is None:
        return
    
    analyzer = CallableAnalyzer(initial_invoke)
    if not analyzer.is_async():
        raise TypeError(
            f'`initial_invoke` should be an `async` function, got {initial_invoke!r}'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 1:
        raise TypeError(
            f'`initial_invoke` should accept `1` parameters, meanwhile the given one expects at '
            f'least `{min_!r}`, got {initial_invoke!r}.'
        )
    
    if min_ != 1:
        if max_ < 1:
            if not analyzer.accepts_positional_parameters():
                raise TypeError(
                    f'`initial_invoke` should accept `1` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got {initial_invoke!r}.'
                )


def validate_get_timeout(get_timeout):
    """
    Validates the given timeout getter.
    
    Parameters
    ----------
    get_timeout : `None`, `callable`
        Timeout getter.
    
    Raises
    ------
    TypeError
        If `get_timeout` is not `None` nor `callable` accepting `1` parameters.
    """
    if get_timeout is None:
        return
    
    analyzer = CallableAnalyzer(get_timeout)
    if analyzer.is_async():
        raise TypeError(
            f'`get_timeout` should not be an `async` function, got {get_timeout!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 1:
        raise TypeError(
            f'`get_timeout` should accept `1` parameter, meanwhile the given one expects at '
            f'least `{min_!r}`, got {get_timeout!r}.'
        )
    
    if min_ != 1:
        if max_ < 1:
            if not analyzer.accepts_positional_parameters():
                raise TypeError(
                    f'`get_timeout` should accept `1` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got {get_timeout!r}.'
                )


def validate_close(close):
    """
    Validates the given closer.
    
    Parameters
    ----------
    close : `callable`
        The closer to validate.
    
    Raises
    ------
    TypeError
        If `close` is not `None` nor an `async-callable` accepting 1 parameter.
    """
    if close is None:
        return
    
    analyzer = CallableAnalyzer(close)
    if not analyzer.is_async():
        raise TypeError(
            f'`close` should be an `async` function, got {close!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 2:
        raise TypeError(
            f'`close` should accept `2` parameters, meanwhile the given one expects at '
            f'least `{min_!r}`, got {close!r}.'
        )
    
    if min_ != 2:
        if max_ < 2:
            if not analyzer.accepts_positional_parameters():
                raise TypeError(
                    f'`close` should accept `2` parameters, meanwhile the given one expects '
                    f'up to `{max_!r}`, got {close!r}.'
                )


def validate_init(init):
    """
    Validates the given initializer.
    
    Parameters
    ----------
    init : `callable`
        The initializer to validate.
    
    Raises
    ------
    TypeError
        If `init` is not `None` or is `async-callable` or accepts less than `3` parameters.
    """
    if (init is None) or (init is object.__init__):
        return
    
    analyzer = CallableAnalyzer(init)
    if analyzer.is_async():
        raise TypeError(
            f'`init` should not be an `async` function, got {init!r}.')
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ < 3:
        raise TypeError(
            f'`init` should accept at least `3` parameters, meanwhile the given one expects at '
            f'least `{min_!r}`, got {init!r}.'
        )


class MenuStructure:
    """
    A special object to store special methods of a menu.
    
    Attributes
    ----------
    check : `None`, `function`
        The function to call when checking whether an event should be called.
        
        Should accept the following parameters:
        
        +-----------+-----------------------+
        | Name      | Type                  |
        +===========+=======================+
        | self      | ``Menu``              |
        +-----------+-----------------------+
        | event     | ``InteractionEvent``  |
        +-----------+-----------------------+
        
        > ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    get_timeout : `None`, `Function`
        Return the time after the menu should be closed.
        
        > Define it as non-positive to never timeout. Not recommended.
        
        Should accept the following parameters:
        
        +-----------+-----------+
        | Name      | Type      |
        +===========+===========+
        | self      | ``Menu``  |
        +-----------+-----------+
        
        
        Should return the following parameter:
        
        +---------------+-----------+
        | Name          | Type      |
        +===============+===========+
        | timeout       | `int`     |
        +---------------+-----------+
    
    close : `None`, `CoroutineFunction`
        Function to call when the menu is closed.
        
        Should accept the following parameters:
        
        +-----------+---------------------------+
        | Name      | Type                      |
        +===========+===========================+
        | self      | ``Menu``                  |
        +-----------+---------------------------+
        | exception | `None`, `BaseException` |
        +-----------+---------------------------+
    
    init : `None`, `Function`
        Initializer function.
    
        Should accept the following parameters:
        
        +-------------------+---------------------------+
        | Name              | Type                      |
        +===================+===========================+
        | self              | ``Menu``                  |
        +-------------------+---------------------------+
        | interaction_event | ``InteractionEvent``      |
        +-------------------+---------------------------+
        | *positional_parameters             | Positional parameters     |
        +-------------------+---------------------------+
        | **keyword_parameters          | Keyword parameters        |
        +-------------------+---------------------------+
    
    initial_invoke : `None`, `CoroutineFunction`
        Function to generate the default page of the menu.
        
        Should accept the following parameters:
        
        +-----------+-----------+
        | Name      | Type      |
        +===========+===========+
        | self      | ``Menu``  |
        +-----------+-----------+
        
    is_final : `bool`
        Whether the Menu has all the required fields fulfilled.
    
    invoke : `None`, `CoroutineFunction`
        The function call for result when invoking the menu.
        
        Should accept the following parameters:
        
        +-----------+-----------------------+
        | Name      | Type                  |
        +===========+=======================+
        | self      | ``Menu``              |
        +-----------+-----------------------+
        | event     | ``InteractionEvent``  |
        +-----------+-----------------------+
        
        Should return the following parameter:
        
        +---------------+-----------+
        | Name          | Type      |
        +===============+===========+
        | should_edit   | `bool`    |
        +---------------+-----------+
    """
    __slots__ = ('check', 'close', 'get_timeout', 'init', 'is_final', 'initial_invoke', 'invoke')
    
    def __new__(cls, class_attributes):
        """
        Creates a new menu structure instance from the given class attributes dictionary.
        
        Parameters
        ----------
        class_attributes : `dict` of (`str`, `object`) items
            Class attributes of a type.
        
        Raises
        ------
        TypeError
            - If `check` is not `None` neither a non-async function.
            - If `check` accepts less or more than `2` parameters.
            - If `invoke` is not `None` nor an `async-callable`.
            - If `invoke` accepts less or more than `2` parameters.
            - If `initial_invoke` is not `None`, nor an `async-callable`.
            - If `initial_invoke` accepts less or more than `1` parameter.
            - If `get_timeout` is not a non-async callable.
            - If `get_timeout` accepts less or more than `1` parameter.
            - If `close` is neither `None` nor `async-callable`.
            - If `close` accepts more or less than `2` parameters.
            - If `init` is not `None` or is `async-callable`.
            - If `init` accepts less than `3` parameters.
        """
        self = class_attributes.get('_menu_structure', None)
        if (self is not None) and (type(self) is cls):
            return self
        
        check = class_attributes.get('check', None)
        validate_check(check)
        
        invoke = class_attributes.get('invoke', None)
        validate_invoke(invoke)
        
        initial_invoke = class_attributes.get('initial_invoke', None)
        validate_initial_invoke(initial_invoke)
        
        get_timeout = class_attributes.get('get_timeout', None)
        validate_get_timeout(get_timeout)
        
        close = class_attributes.get('close', None)
        validate_close(close)
        
        init = class_attributes.get('__init__', None)
        validate_init(init)
        
        if (invoke is None) or (initial_invoke is None):
            is_final = False
        else:
            is_final = True
        
        self = object.__new__(cls)
        self.is_final = is_final
        self.check = check
        self.close = close
        self.init = init
        self.initial_invoke = initial_invoke
        self.invoke = invoke
        self.get_timeout = get_timeout
        
        return self
    
    
    def __repr__(self):
        """Returns the user menu structure's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        check = self.check
        if (check is None):
            field_added = False
        else:
            field_added = True
            
            repr_parts.append(' check=')
            repr_parts.append(repr(check))
        
        close = self.close
        if (close is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' close=')
            repr_parts.append(repr(close))
        
        init = self.init
        if (init is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' init=')
            repr_parts.append(repr(init))
        
        initial_invoke = self.initial_invoke
        if (initial_invoke is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' initial_invoke=')
            repr_parts.append(repr(initial_invoke))
        
        invoke = self.invoke
        if (invoke is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' invoke=')
            repr_parts.append(repr(invoke))
        
        get_timeout = self.get_timeout
        if (get_timeout is not None):
            if field_added:
                repr_parts.append(', ')
            
            repr_parts.append(' get_timeout=')
            repr_parts.append(repr(get_timeout))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def merge(self, other):
        """
        Merges the two menu structure.
        
        Parameters
        ----------
        other : ``MenuStructure``
            The other menu structure to merge into self creating a new one.
        
        Returns
        -------
        new : ``MenuStructure``
        """
        check = self.check
        if check is None:
            check = other.check
        
        close = self.close
        if close is None:
            close = other.close
        
        init = self.init
        if init is None:
            init = other.init
        
        initial_invoke = self.initial_invoke
        if initial_invoke is None:
            initial_invoke = other.initial_invoke
        
        invoke = self.invoke
        if invoke is None:
            invoke = other.invoke
        
        get_timeout = self.get_timeout
        if get_timeout is None:
            get_timeout = other.get_timeout
        
        if (invoke is None) or (initial_invoke is None):
            is_final = False
        else:
            is_final = True
        
        new = object.__new__(type(self))
        new.check = check
        new.close = close
        new.init = init
        new.initial_invoke = initial_invoke
        new.invoke = invoke
        new.get_timeout = get_timeout
        new.is_final = is_final
        
        return new


def _iter_attributes(class_parents, class_attributes):
    """
    Iterates over the given class's attributes and the given attributes.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    class_parents : `tuple` of `type`
        Parent classes.
    class_attributes : `dict` of (`str`, `object`) items
        Class attributes of the source type.
    
    Yields
    ------
    attribute_name : `str`
        An attribute's name.
    attribute_value : `object`
        An attribute's value.
    """
    for class_parent in reversed(class_parents):
        yield from class_parent.__dict__.items()
    
    yield from class_attributes.items()


def _get_component_descriptor(component, component_descriptors, sub_component_descriptors):
    """
    Gets the component descriptor for the given component.
    
    Parameters
    ----------
    component : ``Component``
        The component to get the descriptor for.
    component_descriptors : `dict` of (``ComponentSourceIdentityHasher``, ``ComponentDescriptor``) items
        Already matched component descriptors.
    sub_component_descriptors : `None`, `list` of ``ComponentDescriptor``
        Sub descriptors used by the main one.
    
    Returns
    -------
    component_descriptor : ``ComponentDescriptor``
    """
    component_hasher = ComponentSourceIdentityHasher(component)
    try:
        component_descriptor = component_descriptors[component_hasher]
    except KeyError:
        component_descriptor = ComponentDescriptor(component, sub_component_descriptors)
        component_descriptors[component_hasher] = component_descriptor
    
    return component_descriptor


DISALLOWED_MENU_ATTRIBUTE_NAMES = (
    '__new__',
    '__call__',
    'cancel',
    '_canceller_function',
    '_handle_close_exception',
    'content',
    'embed',
    'allowed_mentions',
    'components',
)

Menu = None

class MenuType(type):
    """
    Meta type for ``Menu``-s.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Creates a Discord entity type. Subclass ``DiscordEntity`` instead of using this class directly as a metaclass.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type`
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `object`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``MenuType``
        """
        if (Menu is not None):
            for attribute_name in DISALLOWED_MENU_ATTRIBUTE_NAMES:
                menu_attribute = getattr(Menu, attribute_name)
                try:
                    actual_attribute = class_attributes[attribute_name]
                except KeyError:
                    class_attributes[attribute_name] = menu_attribute
                else:
                    if (actual_attribute is not menu_attribute):
                        raise TypeError(
                            f'Overwriting `{attribute_name}` is disallowed.'
                        )
        
        old_menu_structure = None
        for class_parent in reversed(class_parents):
            new_menu_structure = MenuStructure(class_parent.__dict__)
            if old_menu_structure is None:
                old_menu_structure = new_menu_structure
            else:
                old_menu_structure = new_menu_structure.merge(old_menu_structure)
        
        new_menu_structure = MenuStructure(class_attributes)
        if (old_menu_structure is not None):
            new_menu_structure = new_menu_structure.merge(old_menu_structure)
        
        class_attributes['_menu_structure'] = new_menu_structure
        
        
        component_descriptors = {}
        components_to_track = []
        new_attributes = {}
        
        for attribute_name, attribute_value in _iter_attributes(class_parents, class_attributes):
            if isinstance(attribute_value, ComponentDescriptor):
                component_hasher = ComponentSourceIdentityHasher(attribute_value._source_component)
                component_descriptors[component_hasher] = attribute_value
                new_attributes[attribute_name] = attribute_value
                continue
            
            if isinstance(attribute_value, Component):
                components_to_track.append((attribute_name, attribute_value))
                continue
        
        
        for attribute_name, attribute_value in components_to_track:
            if attribute_value.components is None:
                sub_component_descriptors = None
            
            else:
                sub_component_descriptors = [
                    _get_component_descriptor(sub_component, component_descriptors, None)
                    for sub_component in attribute_value.iter_components()
                ]
                
            component_descriptor = _get_component_descriptor(
                attribute_value, component_descriptors, sub_component_descriptors
            )
        
            new_attributes[attribute_name] = component_descriptor
            continue    
        
        
        for attribute_name, attribute_value in new_attributes.items():
            class_attributes[attribute_name] = attribute_value
        
        if '__slots__' not in class_attributes:
            class_attributes['__slots__'] = ('__dict__', )
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class Menu(metaclass = MenuType):
    """
    Base class for custom component based menus.
    
    Attributes
    ----------
    _allowed_mentions : ``Ellipsis``, `None`, `list` of `object`
        The used allowed mentions when editing the respective message.
    _canceller : None`, `CoroutineFunction`
        Canceller set as `._canceller_function``, meanwhile the gui is not cancelled.
    _components : `None`, `tuple` of ``Component``
        Rendered components of the menu.
    _component_cache : `dict` of (`int`, ``ComponentProxy``) items
        A dictionary of component proxy identifiers and component proxies.
    _gui_state : `int`
        The gui's state.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +===============================+=======+
        | GUI_STATE_NONE                | 0     |
        +===============================+=======+
        | GUI_STATE_READY               | 1     |
        +-------------------------------+-------+
        | GUI_STATE_EDITING             | 2     |
        +-------------------------------+-------+
        | GUI_STATE_CANCELLING          | 3     |
        +-------------------------------+-------+
        | GUI_STATE_CANCELLED           | 4     |
        +-------------------------------+-------+
        | GUI_STATE_SWITCHING_CONTEXT   | 5     |
        +-------------------------------+-------+
    _timeouter : `None`, ``Timeouter``
        Executes the timeout feature on the menu.
    _tracked_changes : `dict` of (`str`, `object`) items
        The tracked changes by parameter name.
    channel : ``Channel``
        The channel where the menu is executed.
    client : ``Client``
        The executor client instance.
    message : `None`, ``Message``
        The message which executes the menu.
    
    Class Attributes
    ----------------
    _menu_structure : ``MenuStructure``
        Factorized methods added by the user used by the menu itself.
    """
    __slots__ = (
        '_allowed_mentions', '_canceller', '_component_cache', '_gui_state', '_timeouter', '_tracked_changes',
        'channel', 'client', 'message', '_components',
    )
    
    async def __new__(cls, client, target, *positional_parameters, **keyword_parameters):
        """
        Creates a new menu instance.
        
        Parameters
        ----------
        client : ``Client``
            The client instance whi will execute the action.
        target : ``InteractionEvent``, ``Message``, ``Channel``, `int`
            The event to respond to, or the channel to send the message at, or the message to edit.
        *positional_parameters : Positional parameters
            Additional positional parameters.
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Raises
        ------
        TypeError
            If `interaction_event` was not given neither as ``InteractionEvent``, ``Message``, ``Channel``,
             nor as instance.
        RuntimeError
            If `interaction_event` is given as ``InteractionEvent``, but it's client cannot be detected.
        """
        menu_structure = cls._menu_structure
        if not menu_structure.is_final:
            raise RuntimeError(
                f'{cls.__class__.__name__} has not every required fields fulfilled. Required fields: '
                f'`initial_invoke`, `invoke`.'
            )
        
        # use goto
        while True:
            if isinstance(target, InteractionEvent):
                target_message = target.message # Should be `None`
                interaction_event = target
                is_interaction = True
                break
                
            if isinstance(target, Message):
                target_channel_id = target.channel_id
                target_message = target
                is_interaction = False
                break
            
            if isinstance(target, Channel):
                target_channel_id = target.id
                target_message = None
                is_interaction = False
                break
            
            if isinstance(target, int):
                target_channel_id = target
                target_message = None
                is_interaction = False
                break
                
                
            raise TypeError(
                f'`target` can be `{InteractionEvent.__name__}`, '
                f'`{Message.__name__}`, `{Channel.__name__}`, `int`, got '
                f'{target.__class__.__name__}; {target!r}.'
            )
        
        self = object.__new__(cls)
        self._canceller = None
        self.message = target_message
        self.client = client
        self._gui_state = GUI_STATE_NONE
        self._timeouter = None
        self._tracked_changes = {}
        self._allowed_mentions = None
        self._component_cache = {}
        self._components = None
        
        init = menu_structure.init
        if (init is not None):
            init(self, client, target, *positional_parameters, **keyword_parameters)
        
        await menu_structure.initial_invoke(self)
        
        tracked_changes = self._tracked_changes
        if not tracked_changes:
            raise RuntimeError(
                f'{cls.__class__.__name__}\'s `initial_invoke` method: '
                f'{menu_structure.initial_invoke!r} did not change any parameters.'
            )
        
        keyword_parameters = tracked_changes.copy()
        tracked_changes.clear()
        
        allowed_mentions = self._allowed_mentions
        if (allowed_mentions is not None):
            keyword_parameters['allowed_mentions'] = allowed_mentions
        
        keyword_parameters['components'] = self._components
        
        if is_interaction:
            interaction_event_type = interaction_event.type
            if interaction_event_type is INTERACTION_TYPE_APPLICATION_COMMAND:
                if target_message is None:
                    if interaction_event.is_unanswered():
                        await client.interaction_application_command_acknowledge(interaction_event)
                    
                    target_message = await client.interaction_followup_message_create(
                        interaction_event, **keyword_parameters
                    )
                else:
                    await client.interaction_response_message_edit(interaction_event, **keyword_parameters)
            
            elif interaction_event_type is INTERACTION_TYPE_MESSAGE_COMPONENT:
                await client.interaction_component_message_edit(interaction_event, **keyword_parameters)
            
            else:
                # nani desu ka?
                return
        
        else:
            if target_message is None:
                target_message = await client.message_create(target_channel_id, **keyword_parameters)
            else:
                await client.message_edit(target_message, **keyword_parameters)
        
        self.message = target_message
        
        get_timeout = menu_structure.get_timeout
        if (get_timeout is None):
            timeout = -1.0
        else:
            timeout = get_timeout(self)
        
        if (timeout > 0.0):
            self._timeouter = Timeouter(self, timeout)
        
        self._gui_state = GUI_STATE_READY
        self._canceller = cls._canceller_function
        
        client.slasher.add_component_interaction_waiter(target_message, self)
        
        return self
    
    
    async def __call__(self, interaction_event):
        """
        Processes a received component interaction event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        interaction_event : ``InteractionEvent``
            Component interaction event.
        """
        client = self.client
        gui_state = self._gui_state
        if (gui_state == GUI_STATE_EDITING) or (GUI_STATE_EDITING == GUI_STATE_CANCELLING):
            await client.interaction_component_acknowledge(interaction_event)
        
        if (gui_state != GUI_STATE_READY):
            return
        
        check = self._menu_structure.check
        if (check is not None):
            try:
                should_process = check(self, interaction_event)
            except GeneratorExit as err:
                self.cancel(err)
                raise
            
            except BaseException as err:
                self.cancel(err)
                return
            
            if not should_process:
                await client.interaction_component_acknowledge(interaction_event)
                return
        
        self._gui_state = GUI_STATE_EDITING
        
        invoke = self._menu_structure.invoke
        try:
            should_edit = await invoke(self, interaction_event)
        except GeneratorExit as err:
            self.cancel(err)
            raise
        
        except BaseException as err:
            self.cancel(err)
            return
        
        if should_edit:
            tracked_changes = self._tracked_changes
            keyword_parameters = tracked_changes.copy()
            tracked_changes.clear()
            
            allowed_mentions = self._allowed_mentions
            if (allowed_mentions is not None):
                keyword_parameters['allowed_mentions'] = allowed_mentions
            
            try:
                await client.interaction_component_message_edit(
                    interaction_event,
                    **keyword_parameters,
                    components = self._components,
                )
            except GeneratorExit as err:
                self.cancel(err)
                raise
            
            except BaseException as err:
                self.cancel(err)
                return
            
            
            get_timeout = self._menu_structure.get_timeout
            if (get_timeout is None):
                timeout = -1.0
            else:
                timeout = get_timeout(self)
            
            timeouter = self._timeouter
            if (timeout > 0.0):
                if timeouter is None:
                    self._timeouter = Timeouter(self, timeout)
                else:
                    timeouter.set_timeout(timeout)
            else:
                if (timeouter is not None):
                    self._timeouter = None
                    timeouter.cancel()
        
        else:
            try:
                await client.interaction_component_acknowledge(interaction_event)
            except GeneratorExit as err:
                self.cancel(err)
                raise
            
            except BaseException as err:
                self.cancel(err)
                return
        
        if self._gui_state == GUI_STATE_EDITING:
            self._gui_state = GUI_STATE_READY
    
    
    def cancel(self, exception = None):
        """
        Cancels the menu with the given exception.
        
        Parameters
        ----------
        exception : `None`, ``BaseException`` = `None`, Optional
            Exception to cancel the pagination with. Defaults to `None`
        
        Returns
        -------
        canceller_task : `None`, ``Task``
        """
        if self._gui_state in (GUI_STATE_READY, GUI_STATE_EDITING, GUI_STATE_CANCELLING):
            self._gui_state = GUI_STATE_CANCELLED
        
        canceller = self._canceller
        if canceller is None:
            return
        
        self._canceller = None
        
        timeouter = self._timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(KOKORO, canceller(self, exception))
    
    
    async def _canceller_function(self, exception):
        """
        Cancels the gui state, saving the current game if needed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None`, ``BaseException``
        """
        client = self.client
        message = self.message
        
        client.slasher.remove_component_interaction_waiter(message, self)
        
        
        if self._gui_state == GUI_STATE_SWITCHING_CONTEXT:
            # the message is not our, we should not do anything with it.
            return
        
        self._gui_state = GUI_STATE_CANCELLED
        
        close = self._menu_structure.close
        
        try:
            if (close is None):
                handled = await self._handle_close_exception(exception)
            else:
                await close(self, exception)
                handled = True
        except GeneratorExit:
            raise
        
        except BaseException as err:
            await client.events.error(client, f'{self!r}._canceller_function', err)
        
        else:
            if not handled:
                await client.events.error(client, f'{self!r}.cancel', exception)
    
    
    async def _handle_close_exception(self, exception):
        """
        Handles close exception if any.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            The close exception to handle.
        
        Returns
        -------
        handled : `bool`
            Whether the exception was handled.
        """
        if exception is None:
            return True
        
        client = self.client
        message = self.message
        
        if isinstance(exception, CancelledError):
            try:
                await client.message_delete(message)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                if isinstance(err, ConnectionError):
                    # no internet
                    return True
                
                if isinstance(err, DiscordException):
                    if err.code in (
                        ERROR_CODES.unknown_channel, # channel deleted
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.missing_access, # client removed
                    ):
                        return True
                
                await client.events.error(client, f'{self!r}._handle_close_exception', err)
            
            return True
        
        if isinstance(exception, TimeoutError):
            if (message.components is not None):
                try:
                    await client.message_edit(message, components = None)
                except GeneratorExit:
                    raise
                
                except BaseException as err:
                    if isinstance(err, ConnectionError):
                        # no internet
                        return True
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.unknown_channel, # channel deleted
                            ERROR_CODES.missing_access, # client removed
                            ERROR_CODES.missing_permissions, # permissions changed meanwhile
                        ):
                            return True
                    
                    await client.events.error(client, f'{self!r}._handle_close_exception', err)
            
            return True
        
        if isinstance(exception, PermissionError):
            return True
        
        return False
    
    
    @property
    def content(self):
        """
        A get-set-del property to modify the menu's content.
        """
        message = self.message
        if message is None:
            content = None
        else:
            content = message.content
        
        return content
    
    
    @content.setter
    def content(self, content):
        if content is None:
            change_name = 'content'
            change_value = ''
        elif isinstance(content, str):
            change_name = 'content'
            change_value = content
        elif isinstance(content, Embed):
            change_name = 'embed'
            change_value = content
        else:
            change_name = 'content'
            change_value = str(content)
            if len(change_value) > 2000:
                change_value = change_value[:2000]
        
        self._tracked_changes[change_name] = change_value
    
    @content.deleter
    def content(self):
        self._tracked_changes['content'] = None
    
    
    @property
    def embed(self):
        """
        A get-set-del property to modify the menu's embed.
        """
        message = self.message
        if message is None:
            embed = None
        else:
            embeds = message.embeds
            if embeds is None:
                embed = None
            else:
                embed = embeds[0]
        
        return embed
    
    @embed.setter
    def embed(self, embed):
        if (embed is not None) and (not isinstance(embed, Embed)):
            raise TypeError(
                f'`embed` can be `None`, `{Embed.__name__}`, got '
                f'{embed.__class__.__name__}; {embed!r}.'
            )
        
        self._tracked_changes['embed'] =  embed
    
    @embed.deleter
    def embed(self):
        self._tracked_changes['embed'] = None
    
    
    @property
    def allowed_mentions(self):
        """
        A get-set-del property to modify the menu's message's allowed mentions.
        """
        allowed_mentions = self._allowed_mentions
        if allowed_mentions is None:
            allowed_mentions_output = AllowedMentionProxy()
        else:
            allowed_mentions_output = allowed_mentions.copy()
        
        return allowed_mentions_output
    
    @allowed_mentions.setter
    def allowed_mentions(self, allowed_mentions_input):
        if allowed_mentions_input is None:
            allowed_mentions = AllowedMentionProxy()
        elif isinstance(allowed_mentions_input, AllowedMentionProxy):
            allowed_mentions = allowed_mentions_input.copy()
        else:
            allowed_mentions = AllowedMentionProxy(allowed_mentions_input)
        
        self._allowed_mentions = allowed_mentions
    
    @allowed_mentions.deleter
    def allowed_mentions(self):
        self._allowed_mentions = None
    
    
    @property
    def components(self):
        """
        A get-set-del property to modify the menu's components.
        """
        return self._components
    
    
    @components.setter
    def components(self, raw_components):
        components = None
        
        if raw_components is None:
            pass
        
        elif isinstance(raw_components, Component):
            if raw_components.type is ComponentType.row:
                component = raw_components
            else:
                component = create_row(raw_components)
            
            if components is None:
                components = []
            
            components.append(component)
        
        elif isinstance(raw_components, (tuple, list)):
            
            for raw_sub_component in raw_components:
                if isinstance(raw_sub_component, Component):
                    if raw_sub_component.type is ComponentType.row:
                        component = raw_sub_component
                    else:
                        component = create_row(raw_sub_component)
                
                elif isinstance(raw_sub_component, (tuple, list)):
                    component_line = []
                    for raw_sub_sub_component in raw_sub_component:
                        if not isinstance(raw_sub_sub_component, Component):
                            raise TypeError(
                                f'Double nested component can only be `{Component.__name__}`, got '
                                f'{raw_sub_sub_component.__class__.__name__}; {raw_sub_sub_component!r}.'
                            )
                        
                        if raw_sub_sub_component.type is ComponentType.row:
                            raise TypeError(
                                f'Triple nesting components not allowed, got {raw_sub_sub_component!r}.'
                            )
                        
                        component_line.append(raw_sub_sub_component)
                    
                    component = create_row(*component_line)
                
                else:
                    raise TypeError(
                        f'`components` contains an element of unexpected type, got '
                        f'{raw_sub_component.__class__.__name__}; {raw_sub_component!r}; components = {raw_components!r}.'
                    )
                
                if components is None:
                    components = []
                
                components.append(component)
        
        else:
            raise TypeError(
                f'`components` can be `None`, `{Component.__name__}`, (`list`, `tuple`) of repeat, '
                f'no triple nesting, got {raw_components.__class__.__name__}; {raw_components!r}.'
            )
        
        if (components is not None):
            components = tuple(components)
        
        self._components = components
    
    
    @components.deleter
    def components(self):
        self._components = None



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory,
    int_postprocess_parser_factory, nullable_entity_array_parser_factory, nullable_entity_parser_factory,
    nullable_sorted_array_parser_factory, nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    flag_optional_putter_factory, force_string_putter_factory, int_optional_postprocess_putter_factory,
    nullable_entity_array_putter_factory, nullable_entity_optional_putter_factory,
    nullable_string_array_optional_putter_factory, nullable_string_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_entity_array_validator_factory, nullable_entity_validator_factory,
    nullable_string_array_validator_factory, nullable_string_validator_factory, preinstanced_validator_factory,
    url_array_optional_validator_factory, url_optional_validator_factory
)
from ...guild import Guild
from ...user import ClientUserBase, User, ZEROUSER

from ..application_entity import ApplicationEntity
from ..application_executable import ApplicationExecutable
from ..application_install_parameters import ApplicationInstallParameters
from ..eula import EULA
from ..team import Team
from ..third_party_sku import ThirdPartySKU

from .constants import (
    BOT_PUBLIC_DEFAULT, BOT_REQUIRE_CODE_GRANT_DEFAULT, DESCRIPTION_LENGTH_MAX, HOOK_DEFAULT, MAX_PARTICIPANTS_DEFAULT,
    NAME_LENGTH_MAX, NAME_LENGTH_MIN, OVERLAY_COMPATIBILITY_HOOK_DEFAULT, OVERLAY_DEFAULT
)
from .flags import ApplicationFlag
from .preinstanced import ApplicationType

# aliases

parse_aliases = nullable_sorted_array_parser_factory('aliases')
put_aliases_into = nullable_string_array_optional_putter_factory('aliases')
validate_aliases = nullable_string_array_validator_factory('aliases')

# bot_public

parse_bot_public = bool_parser_factory('bot_public', BOT_PUBLIC_DEFAULT)
put_bot_public_into = bool_optional_putter_factory('bot_public', BOT_PUBLIC_DEFAULT)
validate_bot_public = bool_validator_factory('bot_public', BOT_PUBLIC_DEFAULT)

# bot_require_code_grant

parse_bot_require_code_grant = bool_parser_factory('bot_require_code_grant', BOT_REQUIRE_CODE_GRANT_DEFAULT)
put_bot_require_code_grant_into = bool_optional_putter_factory('bot_require_code_grant', BOT_REQUIRE_CODE_GRANT_DEFAULT)
validate_bot_require_code_grant = bool_validator_factory('bot_require_code_grant', BOT_REQUIRE_CODE_GRANT_DEFAULT)

# custom_install_url

parse_custom_install_url = nullable_string_parser_factory('custom_install_url')
put_custom_install_url_into = url_optional_putter_factory('custom_install_url')
validate_custom_install_url = url_optional_validator_factory('custom_install_url')

# deeplink_url

parse_deeplink_url = nullable_string_parser_factory('deeplink_uri')
put_deeplink_url_into = url_optional_putter_factory('deeplink_uri')
validate_deeplink_url = url_optional_validator_factory('deeplink_url')

# description

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# developers

parse_developers = nullable_entity_array_parser_factory('developers', ApplicationEntity)
put_developers_into = nullable_entity_array_putter_factory('developers', ApplicationEntity)
validate_developers = nullable_entity_array_validator_factory('developers', ApplicationEntity)

# eula_id

parse_eula_id = entity_id_parser_factory('eula_id')
put_eula_id_into = entity_id_optional_putter_factory('eula_id')
validate_eula_id = entity_id_validator_factory('eula_id', EULA)

# executables

parse_executables = nullable_entity_array_parser_factory('executables', ApplicationExecutable)
put_executables_into = nullable_entity_array_putter_factory('executables', ApplicationExecutable)
validate_executables = nullable_entity_array_validator_factory('executables', ApplicationExecutable)

# flags

parse_flags = flag_parser_factory('flags', ApplicationFlag)
put_flags_into = flag_optional_putter_factory('flags', ApplicationFlag())
validate_flags = flag_validator_factory('flags', ApplicationFlag)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# hook

parse_hook = bool_parser_factory('hook', HOOK_DEFAULT)
put_hook_into = bool_optional_putter_factory('hook', HOOK_DEFAULT)
validate_hook = bool_validator_factory('hook', HOOK_DEFAULT)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('application_id')

# install_parameters

parse_install_parameters = nullable_entity_parser_factory('install_params', ApplicationInstallParameters)
put_install_parameters_into = nullable_entity_optional_putter_factory('install_params', ApplicationInstallParameters)
validate_install_parameters = nullable_entity_validator_factory('install_parameters', ApplicationInstallParameters)

# max_participants

parse_max_participants = int_postprocess_parser_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,
    (lambda max_participants: MAX_PARTICIPANTS_DEFAULT if max_participants == -1 else max_participants),
)
put_max_participants_into = int_optional_postprocess_putter_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,   
    (lambda max_participants: -1 if max_participants == MAX_PARTICIPANTS_DEFAULT else max_participants),
)
validate_max_participants = int_conditional_validator_factory(
    'max_participants',
    MAX_PARTICIPANTS_DEFAULT,
    (lambda width : width >= 0),
    '>= 0',
)

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# overlay

parse_overlay = bool_parser_factory('overlay', OVERLAY_DEFAULT)
put_overlay_into = bool_optional_putter_factory('overlay', OVERLAY_DEFAULT)
validate_overlay = bool_validator_factory('overlay', OVERLAY_DEFAULT)

# overlay_compatibility_hook

parse_overlay_compatibility_hook = bool_parser_factory('overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT)
put_overlay_compatibility_hook_into = bool_optional_putter_factory(
    'overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT
)
validate_overlay_compatibility_hook = bool_validator_factory(
    'overlay_compatibility_hook', OVERLAY_COMPATIBILITY_HOOK_DEFAULT
)

# owner

def parse_owner(data):
    """
    Parses the application owner.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Application data.
    
    Returns
    -------
    owner : ``ClientUserBase``, ``Team``
    """
    team_data = data.get('team', None)
    if team_data is None:
        user_data = data.get('owner', None)
        if user_data is None:
            owner = ZEROUSER
        else:
            owner = User.from_data(user_data)
    else:
        owner = Team.from_data(team_data)
    
    return owner


def put_owner_into(owner, data, defaults):
    """
    Puts the application's owner data into the given `data` json serializable object.
    
    Parameters
    ----------
    owner : ``ClientUserBase``, ``Team``
        The application's owner.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if owner is ZEROUSER:
        team_data = None
        user_data = None
    
    elif isinstance(owner, ClientUserBase):
        team_data = None
        user_data = owner.to_data(defaults = defaults, include_internals = True)
        
    elif isinstance(owner, Team):
        team_data = owner.to_data(defaults = defaults, include_internals = True)
        user_data = owner.to_data_user()
    
    else:
        team_data = None
        user_data = None
    
    if (team_data is not None) or defaults:
        data['team'] = team_data
    
    if (user_data is not None) or defaults:
        data['owner'] = user_data
    
    return data


def validate_owner(owner):
    """
    Validates the given application owner.
    
    Parameters
    ----------
    owner : `None`, ``ClientUserBase``, ``Team``
        The application's owner.
    
    Returns
    -------
    owner : ``ClientUserBase``, ``Team``
    
    Raises
    ------
    TypeError
        - If `owner`'s type is incorrect.
    """
    if owner is None:
        owner = ZEROUSER
    
    elif not isinstance(owner, (ClientUserBase, Team)):
        raise TypeError(
            f'`owner` can be `None`, `{ClientUserBase.__name__}`, {Team.__name__}, '
            f'got {owner.__class__.__name__}; {owner!r}'
        )
    
    return owner

# primary_sku_id

parse_primary_sku_id = entity_id_parser_factory('primary_sku_id')
put_primary_sku_id_into = entity_id_optional_putter_factory('primary_sku_id')
validate_primary_sku_id = entity_id_validator_factory('primary_sku_id')

# privacy_policy_url

parse_privacy_policy_url = nullable_string_parser_factory('privacy_policy_url')
put_privacy_policy_url_into = url_optional_putter_factory('privacy_policy_url')
validate_privacy_policy_url = url_optional_validator_factory('privacy_policy_url')

# publishers

parse_publishers = nullable_entity_array_parser_factory('publishers', ApplicationEntity)
put_publishers_into = nullable_entity_array_putter_factory('publishers', ApplicationEntity)
validate_publishers = nullable_entity_array_validator_factory('publishers', ApplicationEntity)

# role_connection_verification_url

parse_role_connection_verification_url = nullable_string_parser_factory('role_connections_verification_url')
put_role_connection_verification_url_into = url_optional_putter_factory('role_connections_verification_url')
validate_role_connection_verification_url = url_optional_validator_factory('role_connection_verification_url')

# rpc_origins

parse_rpc_origins = nullable_sorted_array_parser_factory('rpc_origins')
put_rpc_origins_into = nullable_string_array_optional_putter_factory('rpc_origins')
validate_rpc_origins = url_array_optional_validator_factory('rpc_origins')

# slug

parse_slug = nullable_string_parser_factory('slug')
put_slug_into = url_optional_putter_factory('slug')
validate_slug = url_optional_validator_factory('slug')

# tags

parse_tags = nullable_sorted_array_parser_factory('tags')
put_tags_into = nullable_string_array_optional_putter_factory('tags')
validate_tags = nullable_string_array_validator_factory('tags')

# terms_of_service_url

parse_terms_of_service_url = nullable_string_parser_factory('terms_of_service_url')
put_terms_of_service_url_into = url_optional_putter_factory('terms_of_service_url')
validate_terms_of_service_url = url_optional_validator_factory('terms_of_service_url')

# third_party_skus

parse_third_party_skus = nullable_entity_array_parser_factory('third_party_skus', ThirdPartySKU)
put_third_party_skus_into = nullable_entity_array_putter_factory('third_party_skus', ThirdPartySKU)
validate_third_party_skus = nullable_entity_array_validator_factory('third_party_skus', ThirdPartySKU)

# type

parse_type = preinstanced_parser_factory('type', ApplicationType, ApplicationType.none)


def put_type_into(application_type, data, defaults):
    """
    Puts the application's type into the given `data` json serializable object.
    
    Parameters
    ----------
    application_type : ``ApplicationType``
        The application's type.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if application_type is ApplicationType.none:
        value = None
    else:
        value = application_type.value
    
    data['type'] = value
    return data


validate_type = preinstanced_validator_factory('application_type', ApplicationType)

# verify_key

parse_verify_key = nullable_string_parser_factory('verify_key')
put_verify_key_into = nullable_string_putter_factory('verify_key')
validate_verify_key = nullable_string_validator_factory('verify_key', 0, 1024)



from .webhook import *
from .webhook_source_channel import *
from .webhook_source_guild import *


__all__ = (
    *webhook.__all__,
    *webhook_source_channel.__all__,
    *webhook_source_guild.__all__,
)



__all__ = ()

from ...onboarding import OnboardingPromptOption, OnboardingPromptType

from ..audit_log_change import AuditLogChange

from .shared import _convert_preinstanced, convert_nothing, convert_snowflake


def convert_str__name(name, data):
    return convert_nothing('name', data)


def convert_options(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = (*(OnboardingPromptOption.from_data(data) for data in before),)
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = (*(OnboardingPromptOption.from_data(data) for data in after),)
        else:
            after = None
    
    return AuditLogChange('actions', before, after)


def convert_channel_type(name, data):
    return _convert_preinstanced('type', data, OnboardingPromptType)


ONBOARDING_PROMPT_CONVERTERS = {
    'id': convert_snowflake,
    'in_onboarding': convert_nothing,
    'title': convert_str__name,
    'options': convert_options,
    'required': convert_nothing,
    'single_select': convert_nothing,
    'type': convert_channel_type,
}



__all__ = ()


from ...field_parsers import (
    int_parser_factory,
    nullable_string_parser_factory, entity_id_parser_factory
)
from ...field_putters import (
    int_putter_factory,
    entity_id_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, int_conditional_validator_factory,
    nullable_string_validator_factory
)
from .constants import NONCE_LENGTH_MAX
from ..guild import Guild
from ...user import User, ClientUserBase
from scarletio import set_docs
from ....env import CACHE_PRESENCE
from ...core import GUILDS

# chunk_count

parse_chunk_count = int_parser_factory('chunk_count', 0)
put_chunk_count_into = int_putter_factory('chunk_count')
validate_chunk_count = int_conditional_validator_factory(
    'chunk_count',
    0,
    lambda chunk_count : chunk_count >= 0,
    '>= 0',
)

# chunk_index

parse_chunk_index = int_parser_factory('chunk_index', 0)
put_chunk_index_into = int_putter_factory('chunk_index')
validate_chunk_index = int_conditional_validator_factory(
    'chunk_index',
    0,
    lambda chunk_index : chunk_index >= 0,
    '>= 0',
)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# nonce

parse_nonce = nullable_string_parser_factory('nonce')
put_nonce_into = url_optional_putter_factory('nonce')
validate_nonce = nullable_string_validator_factory('nonce', 0, NONCE_LENGTH_MAX)

# users

def parse_users__cache_presence(data, guild_id = 0):
    users = []
    guild = GUILDS.get(guild_id, None)
    guild_profile_datas = data.get('members', None)
    
    if guild is None:
        if (guild_profile_datas is not None):
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id, strong_cache = False)
                users.append(user)
    
    else:
        guild_users = guild.users
        
        if (guild_profile_datas is not None):
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id, strong_cache = False)
                guild_users[user.id] = user
                users.append(user)
        
        presence_datas = data.get('presences', None)
        if (presence_datas is not None):
            for presence_data in presence_datas:
                user_id = int(presence_data['user']['id'])
                try:
                    user = guild_users[user_id]
                except KeyError:
                    pass
                else:
                    user._update_presence(presence_data)
    
    return users


def parse_users__no_cache_presence(data, guild_id = 0):
    users = []
    guild = GUILDS.get(guild_id, None)
    guild_profile_datas = data.get('members', None)
    
    if guild is None:
        if (guild_profile_datas is not None):
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id, strong_cache = False)
                users.append(user)
    else:
        guild_users = guild.users
        
        if (guild_profile_datas is not None):
            for guild_profile_data in guild_profile_datas:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id, strong_cache = False)
                guild_users[user.id] = user
                users.append(user)
    
    return users


if CACHE_PRESENCE:
    parse_users = parse_users__cache_presence
else:
    parse_users = parse_users__no_cache_presence


set_docs(
    parse_users,
    """
    Parses the guild's users from the given data.
    
    > Used when presence caching is enabled.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Guild data.
    users : `dict<int, ClientUserBase>`
        The guild's users.
    guild_id : `int` = `0`, Optional
        The guild's identifier.
        
    Returns
    -------
    users : `list<ClientUserBase>
        Returns `users` parameter.
    """
)


def put_users_into(users, data, defaults, *, guild_id = 0):
    """
    Puts the given `users` into the given `data` json serializable object.
    
    Parameters
    ----------
    users : `list` of ``ClientUserBase``
        Resolved users.
    data : `dict` of (`str`, `object`) items
        Interaction resolved data.
    defaults : `bool`
        Whether default fields values should be included as well.
    guild_id : `int` = `0`, Optional (keyword only)
        The guild's identifier.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    guild_profiles_datas = []
    
    if guild_id:
        for user in users:
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            guild_profile_data = guild_profile.to_data(defaults = defaults, include_internals = True)
            guild_profile_data['user'] = user.to_data(defaults = defaults, include_internals = True)
            guild_profiles_datas.append(guild_profile_data)
            continue
    
    data['members'] = guild_profiles_datas
    
    return data


def validate_users(users):
    """
    Validates the given users.
    
    Parameters
    ----------
    users : `None`, `iterable` of ``ClientUserBase``
        The users to validate.
    
    Returns
    -------
    users : `list` of ``ClientUserBase``
    
    Raises
    ------
    TypeError
        - If `users`'s type is incorrect.
    """
    validated_users = []
    
    if users is None:
        return validated_users
        
    if (getattr(users, '__iter__', None) is None):
        raise TypeError(
            f'`users` can be `None`,`iterable` of `{ClientUserBase.__name__}` elements, '
            f'got {users.__class__.__name__}; {users!r}.'
        )
    
    for user in users:
        if not isinstance(user, ClientUserBase):
            raise TypeError(
                f'`users` elements can be `{ClientUserBase.__name__}`, got {user.__class__.__name__}; '
                f'{user!r}; users = {users!r}.'
            )
        
        validated_users.append(user)
    
    return validated_users



__all__ = ('REGISTERED_COMMANDS', 'REGISTERED_COMMANDS_BY_NAME',)

from os.path import dirname as get_directory_name, join as join_paths

from ... import __package__ as PACKAGE_NAME


COMMANDS = set()

COMMAND_NAME_TO_COMMAND = {}

COMMAND_DIRECTORY = join_paths(get_directory_name(get_directory_name(__file__)), 'commands')
EXTERNAL_IMPORT_ROUTES_FILE = join_paths(get_directory_name(get_directory_name(__file__)), '.external')

COMMAND_IMPORT_ROUTE = (PACKAGE_NAME, 'main', 'commands')

PYTHON_FILE_POSTFIX_NAMES = frozenset(('.py', '.pyd', '.pyc', '.so'))

SYSTEM_DEFAULT_PARAMETER = 'i'


REGISTERED_COMMANDS = set()
REGISTERED_COMMANDS_BY_NAME = {}



__all__ = ('VoiceChannelEffectAnimationType',)

from ...bases import Preinstance as P, PreinstancedBase


class VoiceChannelEffectAnimationType(PreinstancedBase):
    """
    Voice channel effect animation type.
    
    Attributes
    ----------
    name : `str`
        The default name of the voice channel effect animation type.
    value : `int`
        The discord side identifier value of the voice channel effect animation type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``VoiceChannelEffectAnimationType``) items
        Stores the predefined ``VoiceChannelEffectAnimationType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The voice channel effect animation type' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the voice channel effect animation type.
    
    Every predefined voice channel effect animation type can be accessed as class attribute as well:
    
    +-----------------------------------+---------------------------+-------+
    | Class attribute name              | Name                      | Value |
    +===================================+===========================+=======+
    | premium                           | premium                   | 0     |
    +-----------------------------------+---------------------------+-------+
    | basic                             | basic                     | 1     |
    +-----------------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    premium = P(0, 'premium')
    basic = P(1, 'basic')



__all__ = ('ChannelMetadataGuildVoiceBase',)

from scarletio import copy_docs

from .constants import BITRATE_DEFAULT, USER_LIMIT_DEFAULT
from .fields import (
    parse_bitrate, parse_region, parse_user_limit, put_bitrate_into, put_region_into, put_user_limit_into,
    validate_bitrate, validate_region, validate_user_limit
)
from .preinstanced import VoiceRegion

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildVoiceBase(ChannelMetadataGuildMainBase):
    """
    Guild voice channel metadata base.
    
    Attributes
    ----------
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    bitrate : `int`
        The bitrate (in bits) of the voice channel.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites :`None`,  `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    region : ``VoiceRegion``
        The voice region of the channel.
    user_limit : `int`
        The maximal amount of users, who can join the voice channel, or `0` if unlimited.
    
    Class Attributes
    ----------------
    order_group: `int` = `2`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('bitrate', 'region', 'user_limit')
    
    order_group = 2
    
    
    def __new__(
        cls,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
    ):
        """
        Creates a new guild voice base channel metadata from the given parameters.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bitrate
        if bitrate is ...:
            bitrate = BITRATE_DEFAULT
        else:
            bitrate = validate_bitrate(bitrate)
        
        # region
        if region is ...:
            region = VoiceRegion.unknown
        else:
            region = validate_region(region)
        
        # user_limit
        if user_limit is ...:
            user_limit = 0
        else:
            user_limit = validate_user_limit(user_limit)
        
        # Construct
        self = ChannelMetadataGuildMainBase.__new__(
            cls,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        self.bitrate = bitrate
        self.region = region
        self.user_limit = user_limit
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # bitrate
        hash_value ^= self.bitrate << 12
        
        # region
        hash_value ^= hash(self.region)
        
        # user_limit
        hash_value ^= self.user_limit << 15
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # bitrate
        if self.bitrate != other.bitrate:
            return False
        
        # region
        if self.region is not other.region:
            return False
        
        # user_limit
        if self.user_limit != other.user_limit:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildVoiceBase, cls)._create_empty()
        
        self.bitrate = BITRATE_DEFAULT
        self.region = VoiceRegion.unknown
        self.user_limit = USER_LIMIT_DEFAULT
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy)
    def copy(self):
        new = ChannelMetadataGuildMainBase.copy(self)
        new.bitrate = self.bitrate
        new.region = self.region
        new.user_limit = self.user_limit
        return new
    
    
    def copy_with(
        self,
        *,
        bitrate = ...,
        name = ...,
        parent_id = ...,
        permission_overwrites = ...,
        position = ...,
        region = ...,
        user_limit = ...,
    ):
        """
        Copies the guild voice base channel metadata with the given fields.
        
        Parameters
        ----------
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        name : `str`, Optional (Keyword only)
            The channel's name.
        parent_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        permission_overwrites : `None`, `iterable` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        position : `int`, Optional (Keyword only)
            The channel's position.
        region : ``VoiceRegion``, `str`, Optional (Keyword only)
            The voice region of the channel.
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # bitrate
        if bitrate is ...:
            bitrate = self.bitrate
        else:
            bitrate = validate_bitrate(bitrate)
        
        # region
        if region is ...:
            region = self.region
        else:
            region = validate_region(region)
        
        # user_limit
        if user_limit is ...:
            user_limit = self.user_limit
        else:
            user_limit = validate_user_limit(user_limit)
        
        # Construct
        new = ChannelMetadataGuildMainBase.copy_with(
            self,
            name = name,
            permission_overwrites = permission_overwrites,
            parent_id = parent_id,
            position = position,
        )
        new.bitrate = bitrate
        new.region = region
        new.user_limit = user_limit
        return new
    
    
    @copy_docs(ChannelMetadataGuildMainBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            bitrate = keyword_parameters.pop('bitrate', ...),
            name = keyword_parameters.pop('name', ...),
            parent_id = keyword_parameters.pop('parent_id', ...),
            permission_overwrites = keyword_parameters.pop('permission_overwrites', ...),
            position = keyword_parameters.pop('position', ...),
            region = keyword_parameters.pop('region', ...),
            user_limit = keyword_parameters.pop('user_limit', ...),
        )
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # bitrate
        self.bitrate = parse_bitrate(data)
        
        # region
        self.region = parse_region(data)
        
        # user_limit
        self.user_limit = parse_user_limit(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # bitrate
        bitrate = parse_bitrate(data)
        if self.bitrate != bitrate:
            old_attributes['bitrate'] = self.bitrate
            self.bitrate = bitrate
        
        # region
        region = parse_region(data)
        if self.region is not region:
            old_attributes['region'] = self.region
            self.region = region
        
        # user_limit
        user_limit = parse_user_limit(data)
        if self.user_limit != user_limit:
            old_attributes['user_limit'] = self.user_limit
            self.user_limit = user_limit
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # bitrate
        put_bitrate_into(self.bitrate, data, defaults)
        
        # region
        put_region_into(self.region, data, defaults)
        
        # user_limit
        put_user_limit_into(self.user_limit, data, defaults)
        
        return data



__all__ = ()

import warnings

from scarletio import Compound

from ...application import Application
from ...bases import maybe_snowflake
from ...channel import Channel, ChannelType
from ...exceptions import DiscordException, ERROR_CODES
from ...guild import Guild, create_partial_guild_from_id
from ...http import DiscordHTTPClient
from ...invite import Invite, InviteTargetType
from ...permission.permission import PERMISSION_MASK_CREATE_INSTANT_INVITE
from ...user import ClientUserBase

from ..request_helpers import get_channel_id, get_guild_and_id, get_guild_id


def _assert__vanity_invite_edit__vanity_code(vanity_code):
    """
    Asserts the `vanity_code` parameter of ``Client.vanity_invite_edit`` method.
    
    Parameters
    ----------
    vanity_code : `str`
        The new code of the guild's vanity invite.
    
    Raises
    ------
    AssertionError
        - If `vanity_code` is not `str`.
    """
    if not isinstance(vanity_code, str):
        raise AssertionError(
            f'`vanity_code` can be `str`, got {vanity_code.__class__.__name__}; {vanity_code!r}.'
        )
    
    return True


def _assert__invite_create__max_age(max_age):
    """
    Asserts the `max_age` parameter of ``Client.invite_create`` method.
    
    Parameters
    ----------
    max_age : `int`
        After how much time (in seconds) will the invite expire.
    
    Raises
    ------
    AssertionError
        - If `max_age` is not `int`.
    """
    if not isinstance(max_age, int):
        raise AssertionError(
            f'`max_age` can be `int`, got {max_age.__class__.__name__}; {max_age!r}.'
        )
    
    return True


def _assert__invite_create__max_uses(max_uses):
    """
    Asserts the `max_uses` parameter of ``Client.invite_create`` method.
    
    Parameters
    ----------
    max_uses : `int`
        How much times can the invite be used.
    
    Raises
    ------
    AssertionError
        - If `max_uses` is not `int`.
    """
    if not isinstance(max_uses, int):
        raise AssertionError(
            f'`max_uses` can be `int`, got {max_uses.__class__.__name__}; {max_uses!r}.'
        )
    
    return True


def _assert__invite_create__unique(unique):
    """
    Asserts the `unique` parameter of ``Client.invite_create`` method.
    
    Parameters
    ----------
    unique : `bool`
        Whether the created invite should be unique.
    
    Raises
    ------
    AssertionError
        - If `unique` is not `bool`.
    """
    if not isinstance(unique, bool):
        raise AssertionError(
            f'`unique` can be `bool`, got {unique.__class__.__name__}; {unique!r}.'
        )
    
    return True


def _assert__invite_create__temporary(temporary):
    """
    Asserts the `temporary` parameter of ``Client.invite_create`` method.
    
    Parameters
    ----------
    temporary : `bool`
        Whether the invite should give only temporary membership.
    
    Raises
    ------
    AssertionError
        - If `temporary` is not `bool`.
    """
    if not isinstance(temporary, bool):
        raise AssertionError(
            f'`temporary` can be `bool`, got {temporary.__class__.__name__}; {temporary!r}.'
        )
    
    return True


class ClientCompoundInviteEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def vanity_invite_get(self, guild):
        """
        Returns the vanity invite of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's invite will be returned.
        
        Returns
        -------
        invite : `None`, ``Invite``
            The vanity invite of the `guild`, `None` if it has no vanity invite.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        if (guild is None) or guild.partial:
            invite_data = await self.http.vanity_invite_get(guild_id)
            vanity_code = invite_data['code']
        else:
            vanity_code = guild.vanity_code
        
        if vanity_code is None:
            return None
        
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        invite_data = await self.http.invite_get(vanity_code, {})
        return Invite._create_vanity(guild, invite_data)
    
    
    async def vanity_invite_edit(self, guild, vanity_code, *, reason = None):
        """
        Edits the given guild's vanity invite's code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild, what's invite will be edited.
        vanity_code : `str`
            The new code of the guild's vanity invite.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `vanity_code` was not given as `str`.
        """
        guild_id = get_guild_id(guild)
        
        assert _assert__vanity_invite_edit__vanity_code(vanity_code)
        
        await self.http.vanity_invite_edit(guild_id, {'code': vanity_code}, reason)
    
    
    async def invite_create(self, channel, *, max_age=0, max_uses=0, unique=True, temporary=False):
        """
        Creates an invite at the given channel with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel of the created invite.
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `max_age` was not given as `int`.
            - If `max_uses` was not given as `int`.
            - If `unique` was not given as `bool`.
            - If `temporary` was not given as `bool`.
        """
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_invitable() or channel.partial:
                    channel_id = channel.id
                    break
            
            else:
                channel_id = maybe_snowflake(channel)
                if (channel_id is not None):
                    break
            
            raise TypeError(
                f'`channel` can be an invitable channel, `int`'
                f', got {channel.__class__.__name__}; {channel!r}.'
            )
        
        assert _assert__invite_create__max_age(max_age)
        assert _assert__invite_create__max_uses(max_uses)
        assert _assert__invite_create__unique(unique)
        assert _assert__invite_create__temporary(temporary)
        
        data = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
        }
        
        data = await self.http.invite_create(channel_id, data)
        return Invite(data, False)
    
    # 'target_user_id' :
    #     DiscordException BAD REQUEST (400), code=50035: Invalid Form Body
    #     target_user_type.GUILD_INVITE_INVALID_TARGET_USER_TYPE('Invalid target user type')
    # 'target_type', as 0:
    #     DiscordException BAD REQUEST (400), code=50035: Invalid Form Body
    #     target_user_type.BASE_TYPE_CHOICES('Value must be one of (1,).')
    # 'target_type', as 1:
    #     DiscordException BAD REQUEST (400), code=50035: Invalid Form Body
    #     target_user_type.GUILD_INVITE_INVALID_TARGET_USER_TYPE('Invalid target user type')
    # 'target_user_id' and 'target_user_type' together:
    #     DiscordException BAD REQUEST (400), code=50035: Invalid Form Body
    #     target_user_id.GUILD_INVITE_INVALID_STREAMER('The specified user is currently not streaming in this channel')
    # 'target_user_id' and 'target_user_type' with not correct channel:
    #     DiscordException BAD REQUEST (400), code=50035: Invalid Form Body
    #     target_user_id.GUILD_INVITE_INVALID_STREAMER('The specified user is currently not streaming in this channel')
    
    async def stream_invite_create(self, guild, user, *, max_age=0, max_uses=0, unique=True, temporary=False):
        """
        Creates an STREAM invite at the given guild for the specific user. The user must be streaming at the guild,
        when the invite is created.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild where the user streams.
        user : ```ClientUserBase``, `int`
            The streaming user.
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase`` neither as `int`.
            - If `guild` is not ``Guild``.
        ValueError
            - If the user is not streaming at the guild.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if not isinstance(guild, Guild):
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        if isinstance(user, ClientUserBase):
            user_id = user.id
        else:
            user_id = maybe_snowflake(user)
            if user_id is None:
                raise TypeError(
                    f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
                )
        
        try:
            voice_state = guild.voice_states[user_id]
        except KeyError:
            raise ValueError(
                f'The user must stream at a voice channel of the guild. Got user = {user!r}; guild = {guild!r}.'
            ) from None
        
        if not voice_state.self_stream:
            raise ValueError(
                f'The user must stream at a voice channel of the guild. Got user = {user!r}; guild = {guild!r}.'
            )
        
        assert _assert__invite_create__max_age(max_age)
        assert _assert__invite_create__max_uses(max_uses)
        assert _assert__invite_create__unique(unique)
        assert _assert__invite_create__temporary(temporary)
        
        data = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
            'target_user_id': user_id,
            'target_type': InviteTargetType.stream.value,
        }
        
        data = await self.http.invite_create(voice_state.channel.id, data)
        return Invite(data, False)
    
    # Could not find correct application:
    #    DiscordException Bad Request (400), code=50035: Invalid Form Body
    #    target_application_id.GUILD_INVITE_INVALID_APPLICATION('The specified application is not embedded')
    
    async def application_invite_create(
        self, channel, application, *, max_age=0, max_uses=0, unique=True, temporary=False
    ):
        """
        Creates an EMBEDDED_APPLICATION invite to the specified voice channel. The application must have must have
        `embedded` flag.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int
            The target channel of the invite.
        application : ``Application``, `int`
            The embedded application to open in the voice channel.
            
            > The application must have `EMBEDDED_APPLICATION` flag.
        
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire.
            
            > If given as `0` (so by default) then the created invite will never expire.
        
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used.
            
            > If given as `0` (so by default) then the created invite will have no use limit.
        
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel``, neither as `int`.
            - If `application` was not given neither as ``Application`` nor as`int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_guild_voice)
        
        if isinstance(application, Application):
            application_id = application.id
        else:
            application_id = maybe_snowflake(application)
            if application_id is None:
                raise TypeError(
                    f'`application` can be `{Application.__name__}`, `int`, got '
                    f'{application.__class__.__name__}; {application!r}.'
                )
        
        assert _assert__invite_create__max_age(max_age)
        assert _assert__invite_create__max_uses(max_uses)
        assert _assert__invite_create__unique(unique)
        assert _assert__invite_create__temporary(temporary)
        
        data = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
            'target_application_id': application_id,
            'target_type': InviteTargetType.embedded_application.value,
        }
        
        data = await self.http.invite_create(channel_id, data)
        return Invite(data, False)
    
    
    async def invite_create_preferred(self, guild, **kwargs):
        """
        Creates an invite to the guild's preferred channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild . ``Guild``
            The guild to her the invite will be created to.
        **kwargs : Keyword parameters
            Additional keyword parameters to describe the created invite.
        
        Other Parameters
        ----------------
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - if `guild` is not ``Guild`˛
        ValueError
            If the guild has no channel to create invite to.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if not isinstance(guild, Guild):
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        while True:
            if not guild.channels:
                raise ValueError(
                    f'The guild has no channels (yet?), try waiting for dispatch or create a channel. '
                    f'Got guild={guild!r}.'
                )

            channel = guild.system_channel
            if channel is not None:
                break
            
            channel = guild.widget_channel
            if channel is not None:
                break
            
            for channel_type in (ChannelType.guild_text, ChannelType.guild_voice):
                for channel in guild.channels.values():
                    if channel.type == ChannelType.guild_category:
                        for channel in channel.channels:
                            if channel.type == channel_type:
                                break
                    if channel.type == channel_type:
                        break
                if channel.type == channel_type:
                    break
            else:
                raise ValueError(
                    f'The guild has only category channels and cannot create invite from them. '
                    f'Got guild={guild!r}.'
                )
            break
        
        # Check permission, because it can save a lot of time >.>
        if not channel.cached_permissions_for(self) & PERMISSION_MASK_CREATE_INSTANT_INVITE:
            return None
        
        try:
            return (await self.invite_create(channel, **kwargs))
        except DiscordException as err:
            if err.code in (
                ERROR_CODES.unknown_channel, # the channel was deleted meanwhile
                ERROR_CODES.missing_permissions, # permissions changed meanwhile
                ERROR_CODES.missing_access, # client removed
            ):
                return None
            raise
    
    
    async def invite_get(self, invite, *, with_count=...):
        """
        Requests a partial invite with the given code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``, `str`
            The invites code.
        
        with_count : `bool`, Optional (Keyword only)
            Whether the invite should contain the respective guild's user and online user count.
            
            Defaults to `True`.
            
            Deprecated and will be removed in 2023 January.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `invite_code` was not given as `str`.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {invite.__class__.__name__}; '
                f'{invite!r}.'
            )
        
        if (with_count is not ...):
            warnings.warn(
                (
                    f'`with_count` parameter of `{self.__class__.__name__}` is deprecated and will be removed in '
                    f'2023 January. The parameter is always defaulting to `True`.',
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        
        invite_data = await self.http.invite_get(invite_code, {'with_counts': True})
        
        if invite is None:
            invite = Invite(invite_data, False)
        else:
            if invite.partial:
                updater = Invite._update_attributes
            else:
                updater = Invite._update_counts_only
            
            updater(invite, invite_data)
        
        return invite
    
    
    async def invite_get_all_guild(self, guild):
        """
        Gets the invites of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        invite_datas = await self.http.invite_get_all_guild(guild_id)
        return [Invite(invite_data, False) for invite_data in invite_datas]
    
    
    async def invite_get_all_channel(self, channel):
        """
        Gets the invites of the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_invitable() or channel.partial:
                    channel_id = channel.id
                    break
            
            else:
                channel_id = maybe_snowflake(channel)
                if (channel_id is not None):
                    break
            
            raise TypeError(
                f'`channel` can be an invitable ``Channel``, `int`, got {channel.__class__.__name__}; {channel!r}.'
            )
        
        invite_datas = await self.http.invite_get_all_channel(channel_id)
        return [Invite(invite_data, False) for invite_data in invite_datas]
    
    
    async def invite_delete(self, invite, *, reason = None):
        """
        Deletes the given invite.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``
            The invite to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {invite.__class__.__name__}; {invite!r}.'
            )
        
        invite_data = await self.http.invite_delete(invite_code, reason)
        
        if invite is None:
            invite = Invite(invite_data, False)
        else:
            invite._update_attributes(invite_data)
        
        return invite



__all__ = ('ApplicationEntity',)

from ...bases import DiscordEntity

from .fields import parse_id, parse_name, put_id_into, put_name_into, validate_id, validate_name


class ApplicationEntity(DiscordEntity):
    """
    An un-typed entity stored inside of an ``Application``, as one of it's `.developers`, `.publishers`.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the entity.
    name : `str`
        The name of the entity.
    """
    __slots__ = ('name', )
    
    
    def __new__(cls, *, name = ...):
        """
        Creates a new partial application entity.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The name of the entity.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        self = object.__new__(cls)
        self.id = 0
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new application entity.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Developers or Publisher data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.name = parse_name(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the application entity into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        if include_internals:
            put_id_into(self.id, data, defaults)
        
        put_name_into(self.name, data, defaults)
        
        return data
    
    
    @classmethod
    def _create_empty(cls, entity_id):
        """
        Creates a new application entity instance with it's attribute set to their default values.
        
        Parameters
        ----------
        entity_id : `int`
            The entity's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = entity_id
        self.name = ''
        return self
    

    @classmethod
    def precreate(cls, entity_id, **keyword_parameters):
        """
        Creates an application entity instance. Since these objects are not cached, the only advantage of using
        ``.precreate`` is that it allows setting ``.id``.
        
        Parameters
        ----------
        entity_id : `int`
            The entity's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The entity's name.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        entity_id = validate_id(entity_id)

        if keyword_parameters:
            processable = []
            
            try:
                name = keyword_parameters.pop('name')
            except KeyError:
                pass
            else:
                name = validate_name(name)
                processable.append(('name', name))
            
            if keyword_parameters:
                raise TypeError(
                    f'Unused or unsettable keyword parameters: {keyword_parameters!r}.'
                )
        
        else:
            processable = None
        
        
        self = cls._create_empty(entity_id)
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    def __repr__(self):
        """Returns the entity's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        entity_id = self.id
        if entity_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application entities are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two application entities are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the application entity."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application entity.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = 0
        new.name = self.name
        return new
    
    
    def copy_with(self, *, name = ...):
        """
        Copies the application entity with the defined fields.
        
        Parameters
        ----------
        name : `str`, Optional (Keyword only)
            The entity's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        new = object.__new__(type(self))
        new.id = 0
        new.name = name
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the entity is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)



__all__ = ('AutoModerationActionType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..action_metadata import (
    AutoModerationActionMetadataBase, AutoModerationActionMetadataBlock, AutoModerationActionMetadataSendAlertMessage,
    AutoModerationActionMetadataTimeout
)


class AutoModerationActionType(PreinstancedBase):
    """
    Represents an ``AutoModerationAction``'s type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the auto moderation action type.
    name : `str`
        The default name of the auto moderation action type.
    metadata_type : ``AutoModerationActionMetadataBase``
        The action type's respective metadata type.
    
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``AutoModerationActionType``) items
        Stores the predefined auto moderation action types. This container is accessed when translating a Discord side
        identifier of a auto moderation action type. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `str`
        The auto moderation action types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the auto moderation action types.
    
    Every predefined auto moderation action type is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | Class attribute name  | Name                  | Value     | Metadata type                                     | Description                                                           |
    +=======================+=======================+===========+===================================================+=======================================================================+
    | none                  | none                  | 0         | ``AutoModerationActionMetadataBase``              | N/A                                                                   |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | block_message         | block message         | 1         | ``AutoModerationActionMetadataBlock``             | Blocks the message's content according to the rule.                   |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | send_alert_message    | send alert message    | 2         | ``AutoModerationActionMetadataSendAlertMessage``  | Sends an alert message to the specified channel.                      |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    | timeout               | timeout               | 3         | ``AutoModerationActionMetadataTimeout``           | Timeouts the user. Only applicable for `keyword` rules. Max 4 weeks.  |
    +-----------------------+-----------------------+-----------+---------------------------------------------------+-----------------------------------------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new auto moderation action type with the given value.
        
        Parameters
        ----------
        value : `int`
            The auto moderation action type's identifier value.
        
        Returns
        -------
        self : ``AutoModerationActionType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.metadata_type = AutoModerationActionMetadataBase
        
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates an ``AutoModerationActionType`` and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the auto moderation action type.
        name : `str`
            The default name of the auto moderation action type.
        metadata_type : ``AutoModerationRuleTriggerMetadataBase``
            The action type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    # predefined
    none = P(0, 'none', AutoModerationActionMetadataBase)
    block_message = P(1, 'block message', AutoModerationActionMetadataBlock)
    send_alert_message = P(2, 'send alert message', AutoModerationActionMetadataSendAlertMessage)
    timeout = P(3, 'timeout', AutoModerationActionMetadataTimeout)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .constants import PLUGIN_ACTION_FLAG_NAME_LOOKUP
from .exceptions import PluginError
from .helpers import _iter_lookup_plugin_names_and_paths, _try_get_plugin
from .plugin import Plugin
from .plugin_tree import PluginTree


class PluginExtractor(RichAttributeErrorBaseType):
    """
    Extracts different plugin related values from related input(s).
    
    Attributes
    ----------
    input_plugin_trees : `None`, `set` of ``PluginTree``
        Extracted plugin trees from the input.
    input_plugins : `None`, `set` of ``Plugin``
        Extracted plugins from the input.
    input_strings : `None`, `set` of `str`
        Extracted plugins from the input.
    name : `str`
        The extracted value(s) name.
    """
    __slots__ = ('input_plugin_trees', 'input_plugins', 'input_strings', 'name')
    
    def __new__(cls, name, value):
        """
        Creates a new plugin extractor instance from the given value(s).
        
        Parameters
        ----------
        name : `str`
            The extracted value(s) name.
        value : `None`, `str`, ``Plugin``, ``PluginTree`` `iterable` of (`None`, `str`, ``Plugin``, ``PluginTree``)
            The value(s) to extract.
        
        Raises
        ------
        TypeError
            - If `value`'s type is invalid.
        """
        input_strings = None
        input_plugins = None
        input_plugin_trees = None
        
        if value is None:
            pass
        
        elif isinstance(value, str):
            input_strings = {str(value)}
        
        elif isinstance(value, Plugin):
            input_plugins = {value}
        
        elif isinstance(value, PluginTree):
            input_plugin_trees = {value}
        
        else:
            if getattr(value, '__iter__', None) is None:
                raise TypeError(
                    f'`{name}` can be either `None`, `str`, `{Plugin.__name__}`, `{PluginTree.__name__}`, `iterable`, '
                    f'got {value.__class__.__name__}; {value!r}.'
                )
            
            for iterated_value in value:
                if iterated_value is None:
                    pass
                
                elif isinstance(iterated_value, str):
                    if input_strings is None:
                        input_strings = set()
                    
                    input_strings.add(iterated_value)
                
                elif isinstance(iterated_value, Plugin):
                    if input_plugins is None:
                        input_plugins = set()
                    
                    input_plugins.add(iterated_value)
                
                elif isinstance(iterated_value, PluginTree):
                    if input_plugin_trees is None:
                        input_plugin_trees = set()
                    
                    input_plugin_trees.add(iterated_value)
                
                else:
                    raise TypeError(
                        f'`{name}` elements can be `None`, `str`, `{Plugin.__name__}`, `{PluginTree.__name__}`, '
                        f'got {iterated_value.__class__.__name__}; {iterated_value!r}; value = {value!r}.'
                    )
        
        self = object.__new__(cls)
        self.input_plugin_trees = input_plugin_trees
        self.input_plugins = input_plugins
        self.input_strings = input_strings
        self.name = name
        return self
    
    
    @classmethod
    def from_strings(cls, name, value):
        """
        Creates a new plugin extractor instance from the given string(s).
        
        Parameters
        ----------
        name : `str`
            The extracted value(s) name.
        value : `str`, `iterable` of `str`
            The value(s) to extract.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If `value`'s type is invalid.
        """
        input_strings = None
        
        if isinstance(value, str):
            input_strings = {str(value)}
        else:
            if getattr(value, '__iter__', None) is None:
                raise TypeError(
                    f'`{name}` can be either `str`, `iterable`, got {value.__class__.__name__}; {value!r}.'
                )
            
            for iterated_value in value:
                if isinstance(iterated_value, str):
                    if input_strings is None:
                        input_strings = set()
                    
                    input_strings.add(iterated_value)
                else:
                    raise TypeError(
                        f'`{name}` elements can be `None`, `str`, `{Plugin.__name__}`, `{PluginTree.__name__}`, '
                        f'got {iterated_value.__class__.__name__}; {iterated_value!r}; value = {value!r}.'
                    )

        self = object.__new__(cls)
        self.input_plugin_trees = None
        self.input_plugins = None
        self.input_strings = input_strings
        self.name = name
        return self
    
    
    def __repr__(self):
        """Returns the representation of the plugin extractor."""
        repr_parts = ['<', self.__class__.__name__]
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        input_plugin_trees = self.input_plugin_trees
        if (input_plugin_trees is not None):
            repr_parts.append(', input_plugin_trees = ')
            repr_parts.append(repr(input_plugin_trees))
        
        input_plugins = self.input_plugins
        if (input_plugins is not None):
            repr_parts.append(', input_plugins = ')
            repr_parts.append(repr(input_plugins))
        
        input_strings = self.input_strings
        if (input_strings is not None):
            repr_parts.append(', input_strings = ')
            repr_parts.append(repr(input_strings))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def iter_input_strings(self):
        """
        Iterates over the input strings of the plugin extractor.
        
        This method is an iterable generator.
        
        Yields
        ------
        input_string : `str`
        """
        input_strings = self.input_strings
        if (input_strings is not None):
            yield from input_strings
    
    
    def iter_input_plugins(self):
        """
        Iterates over the input plugins of the plugin extractor.
        
        This method is an iterable generator.
        
        Yields
        ------
        input_plugin : ``Plugin``
        """
        input_plugins = self.input_plugins
        if (input_plugins is not None):
            yield from input_plugins
    
    
    def iter_input_plugin_trees(self):
        """
        Iterates over the input plugin trees of the plugin extractor.
        
        This method is an iterable generator.
        
        Yields
        ------
        input_plugin_tree : ``PluginTree``
        """
        input_plugin_trees = self.input_plugin_trees
        if (input_plugin_trees is not None):
            yield from input_plugin_trees
    
    
    def get_plugin_names_and_paths(self, *, register_directories_as_roots = False):
        """
        Gets plugin names and paths from the input.
        
        Parameters
        ----------
        register_directories_as_roots : `bool` = `False`, Optional (Keyword only)
            Whether directory roots should be registered.
        
        Returns
        -------
        plugin_names_and_paths : `set` of `tuple` ((`str`, `None`), `str`)
            Plugin name - plugin path pairs.
        """
        return {
            *self.iter_plugin_names_and_paths_from_strings(register_directories_as_roots),
            *self.iter_plugin_names_and_paths_from_plugins(),
            *self.iter_plugin_names_and_paths_from_plugin_trees(),
        }
    
    
    def get_plugins(self, *, raise_if_cannot_resolve = True):
        """
        Gets the plugins from the input.
        
        Returns
        -------
        plugins : `set`
            Plugin name - plugin path pairs.
        
        Parameters
        ----------
        raise_if_cannot_resolve : `bool` = `True`, Optional (Keyword only)
            Whether exception should be raised if a plugin cannot be resolved.
        
        Raises
        ------
        PluginError
            If a plugin could not be found.
        """
        return {
            *self.iter_plugins_from_strings(raise_if_cannot_resolve),
            *self.iter_input_plugins(),
            *self.iter_plugin_from_plugin_trees(),
        }
    
    
    def iter_plugin_names_and_paths_from_strings(self, register_directories_as_roots):
        """
        gets the plugin names and the paths from the input strings.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        register_directories_as_roots : `bool`
            Whether directory roots should be registered.
        
        Yields
        ------
        plugin_name : `None`, `str`
            Plugin's  name.
        plugin_path : `str`
            Path of the plugin file.
        """
        for input_string in self.iter_input_strings():
            yield from _iter_lookup_plugin_names_and_paths(input_string, register_directories_as_roots, False)
    
    
    def iter_plugin_names_and_paths_from_plugins(self):
        """
        Gets the plugin names and the paths from the input plugins.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin_name : `str`
            Plugin's  name.
        plugin_path : `str`
            Path of the plugin file.
        """
        for input_plugin in self.iter_input_plugins():
            yield input_plugin.name, input_plugin.path
    
    
    def iter_plugin_names_and_paths_from_plugin_trees(self):
        """
        Gets the plugin names and the paths from the input plugin trees.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin_name : `str`
            Plugin's  name.
        plugin_path : `str`
            Path of the plugin file.
        """
        for plugin in self.iter_plugin_from_plugin_trees():
            yield plugin.name, plugin.path
    
    
    def iter_plugins_from_strings(self, raise_if_cannot_resolve):
        """
        Gets the plugins from the input strings.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        raise_if_cannot_resolve : `bool`
            Whether exception should be raised if a plugin cannot be resolved.
        
        Yields
        ------
        plugin : ``Plugin``
        
        Raises
        ------
        PluginError
            If a plugin could not be found.
        """
        for plugin_name, plugin_path in self.iter_plugin_names_and_paths_from_strings(False):
            plugin = _try_get_plugin(plugin_name, plugin_path)
            if (plugin is not None):
                yield plugin
                continue
            
            if raise_if_cannot_resolve:
                raise PluginError(
                    f'No plugin was added with name: `{plugin_name}`.',
                    action = PLUGIN_ACTION_FLAG_NAME_LOOKUP,
                    value = (plugin_name, plugin_path),
                )
    
    
    def iter_plugin_from_plugin_trees(self):
        """
        Gets the plugins from the input plugins trees.
        
        This method is an iterable generator.
        
        Yields
        ------
        plugin : ``Plugin``
        """
        for input_plugin_tree in self.iter_input_plugin_trees():
            yield from input_plugin_tree.iter_plugins()



from .constants import *
from .fields import *
from .flags import *
from .guild_profile import *


__all__ = (
    *constants.__all__,
    *fields.__all__,
    *flags.__all__,
    *guild_profile.__all__,
)



__all__ = ('SolarClient', )

from random import choice

from scarletio import RichAttributeErrorBaseType, Task, TaskGroup, WeakReferer, run_coroutine, to_json
from scarletio.web_common.headers import AUTHORIZATION, CONTENT_TYPE

from ...discord.channel import Channel
from ...discord.client import Client
from ...discord.client.request_helpers import get_channel_guild_id_and_id
from ...discord.core import KOKORO
from ...discord.voice.utils import try_get_voice_region

from .event_handler_plugin import SolarLinkEventManager
from .exceptions import SolarAuthenticationError
from .node import SolarNode
from .player import SolarPlayer
from .player_base import SolarPlayerBase
from .route_planner import get_route_planner
from .track import GetTracksResult, Track


class SolarClient(RichAttributeErrorBaseType):
    """
    Represents a lavalink client used to manage nodes and connections.
    
    Attributes
    ----------
    _client_reference : ``Weakreferer`` to ``Client``
        Weakreference to the extended client instance.
    _events : ``SolarLinkEventManager``
        Event plugin for solarlink specific events.
    _player_queue : `None`, `list` of ``SolarPlayerBase``
        Solar players to join back to a node.
    nodes : `set` of ``SolarNode``
        All nodes the client is connected to.
    players : `dict` of (`int`, ``SolarPlayerBase``) items
        Active players of the client by their guild's identifier as key.
    """
    __slots__ = ('_client_reference', '_events', '_player_queue', 'nodes', 'players')
    
    def __new__(cls, client):
        """
        Creates and binds a lavalink manager client.
        
        Parameters
        ----------
        client : ``Client``
            Hata client instance to extend.
        """
        if not isinstance(client, Client):
            raise TypeError(
                f'`client` parameter can be `{Client.__name__}`, got {client.__class__.__name__}; {client!r}.'
            )
        
        event_plugin = SolarLinkEventManager()
        client.events.register_plugin(event_plugin)
        
        client_reference = WeakReferer(client)
        
        self = object.__new__(cls)
        self._client_reference = client_reference
        self.nodes = set()
        self.players = {}
        self._player_queue = None
        self._events = event_plugin
        return self
    
    
    def add_node(self, host, port, password, region, resume_key = None, reconnect_attempts=3):
        """
        Adds a node to Lavalink's node manager.
        
        The return of the method depends on the thread, from which it was called from.
        
        Parameters
        ----------
        host : `str`
            The address of the lavalink node.
        port : `int`
            The port used by the lavalink node.
        password : `str`
            The password used for authentication.
        region : `str`
            The region to assign this node to.
        resume_key : `str` = `None`, Optional
            A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
        reconnect_attempts : `int` = `3`, Optional
            The amount of times connection with the node will be reattempted before giving up.
            Set to `-1` for infinite.
        
        Returns
        -------
        task : `bool`, ``Task``, ``FutureAsyncWrapper``
            - If the method was called from the client's thread (KOKORO), then returns a ``Task``. The task will return
                `True`, if connecting was successful.
            - If the method was called from an ``EventThread``, but not from the client's, then returns a
                `FutureAsyncWrapper`. The task will return `True`, if connecting was successful.
            - If the method was called from any other thread, then waits for the connector task to finish and returns
                `True`, if it was successful.
        
        Raises
        ------
        RuntimeError
            If the ``SolarClient``'s client is already deconstructed.
        TypeError
            - If `host` is not `str`.
            - If `port` is not `int`.
            - If `password is not `str`.
            - If `region` is neither `None`, nor ``VoiceRegion``.
            - If `resume_key` is neither `None`, nor `str`.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        node = SolarNode(
            client,
            host,
            port,
            password,
            region,
            resume_key,
            reconnect_attempts,
        )
        
        self.nodes.add(node)
        
        return run_coroutine(node.start(), KOKORO)
    
    
    async def get_tracks(self, query):
        """
        Gets all tracks associated with the given query.
        
        Parameters
        ----------
        query: : `str`
            The query to perform a search for.
        
        Returns
        -------
        tracks : `None`, ``GetTracksResult``
            Decoded tracks.
        
        Raises
        ------
        RuntimeError
            - If there are no available nodes.
            - If the ``SolarClient``'s client is already deconstructed.
        SolarAuthenticationError
            Authentication failed towards the node.
        """
        available_nodes = self.available_nodes
        if not available_nodes:
            raise RuntimeError('No available nodes!')
        
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        node = choice(self.available_nodes)
        
        async with client.http.get(
            f'http://{node._host}:{node._port}/loadtracks',
            headers = {
                AUTHORIZATION: node._password,
            },
            params = {
                'identifier': query,
            },
        ) as response:
            if response.status == 200:
                data = await response.json()
            
            elif response.status in (401, 403):
                raise SolarAuthenticationError(node, response)
            
            else:
                data = None
        
        if data is None:
            result = None
        else:
            result = GetTracksResult(data)
        
        return result
    
    
    async def decode_track(self, track):
        """
        Decodes a base64-encoded track string into a dictionary.
        
        Parameters
        ----------
        track : `str`
            The base64-encoded track string.
        
        Returns
        -------
        track : `None`, ``Track``
            Decoded track data.
        
        Raises
        ------
        RuntimeError
            - If there are no available nodes.
            - If the ``SolarClient``'s client is already deconstructed.
        SolarAuthenticationError
            Authentication failed towards the node.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        available_nodes = self.available_nodes
        if not available_nodes:
            raise RuntimeError('No available nodes!')
        
        node = choice(self.available_nodes)
        
        async with client.http.get(
            f'http://{node._host}:{node._port}/decodetrack',
            headers = {
                AUTHORIZATION: node._password,
            },
            params = {
                'track': track
            },
        ) as response:
            if response.status == 200:
                track_data = await response.json()
            
            elif response.status == 401 or response.status == 403:
                raise SolarAuthenticationError(node, response)
            
            else:
                track_data = None
        
        if track_data is None:
            track = None
        else:
            track = Track(track_data)
        
        return track
    
    
    async def decode_tracks(self, tracks):
        """
        Decodes a list of base64-encoded track strings.
        
        Parameters
        ----------
        tracks : `list` of `str`
            A list of base64-encoded track strings.
        
        Returns
        -------
        tracks : `list` of ``Track``
            The decoded tracks.
        
        Raises
        ------
        RuntimeError
            - If there are no available nodes.
            - If the ``SolarClient``'s client is already deconstructed.
        SolarAuthenticationError
            Authentication failed towards the node.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        available_nodes = self.available_nodes
        if not available_nodes:
            raise RuntimeError('No available nodes!')
        
        node = choice(available_nodes)
        
        async with client.http.post(
            f'http://{node._host}:{node._port}/decodetracks',
            headers = {
                AUTHORIZATION: node._password,
                CONTENT_TYPE: 'application/json',
            },
            data = to_json(tracks),
        ) as response:
            if response.status == 200:
                track_datas = await response.json()
            
            elif response.status in (401, 403):
                raise SolarAuthenticationError(node, response)
            
            else:
                track_datas = None
        
        if (track_datas is None) or (not track_datas):
            tracks = []
        else:
            tracks = [Track(track_data) for track_data in track_datas]
        
        return tracks
    
    
    async def routeplanner_status(self, node):
        """
        Gets the routeplanner status of the target node.
        
        This method is a coroutine.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node to use for the query.
        
        Returns
        -------
        route_planner : `None`, ``RoutePlannerBase``
        
        Raises
        ------
        RuntimeError
            If the ``SolarClient``'s client is already deconstructed.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        async with client.http.get(
            f'http://{node._host}:{node._port}/routeplanner/status',
            headers = {
                AUTHORIZATION: node._password,
            },
        ) as response:
            if response.status == 200:
                route_planner_data = await response.json()
            
            elif response.status in (401, 403):
                raise SolarAuthenticationError(node, response)
            
            else:
                route_planner_data = None
        
        if route_planner_data is None:
            route_planner = None
        else:
            route_planner = get_route_planner(route_planner_data)
        
        return route_planner
        
    
    async def routeplanner_free_address(self, node, address):
        """
        Removes a single address from the addresses which are currently marked as failing.
        
        This method is a coroutine.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node to use for the query.
        address : `str`
            The address to free.
        
        Returns
        -------
        success : `bool`
            Whether the address is freed up.
        
        Raises
        ------
        SolarAuthenticationError
            Authentication failed towards the node.
        RuntimeError
            If the ``SolarClient``'s client is already deconstructed.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        async with client.http.post(
            f'http://{node._host}:{node._port}/routeplanner/free/address',
            headers = {
                AUTHORIZATION: node._password,
                CONTENT_TYPE: 'application/json',
            },
            data = to_json(
                {'address': address}
            ),
        ) as response:
            status = response.status
            if status in (401, 403):
                raise SolarAuthenticationError(node, response)
            
            if status == 204:
                # Success
                return True
            
            if status == 500:
                # The node has no routeplanner configured.
                return False
            
            # Unexpected case.
            return False
    
    
    async def routeplanner_free_address_all(self, node):
        """
        Removes all addresses from the list which holds the addresses which are marked failing.
        
        This method is a coroutine.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node to use for the query.
        
        Returns
        -------
        success : `bool`
        
        Raises
        ------
        SolarAuthenticationError
            Authentication failed towards the node.
        RuntimeError
            If the ``SolarClient``'s client is already deconstructed.
        """
        client = self._client_reference()
        if client is None:
            raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
        
        async with client.http.post(
            f'http://{node._host}:{node._port}/routeplanner/free/all',
            headers = {
                AUTHORIZATION: node._password,
            },
        ) as response:
            status = response.status
            if status in (401, 403):
                raise SolarAuthenticationError(node, response)
            
            if status == 204:
                # Success
                return True
            
            if status == 500:
                # The node has no routeplanner configured
                return False
            
            # Unexpected case.
            return False
    
    
    @property
    def available_nodes(self):
        """
        Returns a list of available nodes.
        
        Returns
        -------
        available_nodes : `list` of ``SolarNode``
        """
        return [node for node in self.nodes if node.available]
    
    
    def remove_node(self, node):
        """
        Removes a node.
        
        Parameters
        ----------
        node : ``SolarNode``
            The node to remove from the list.
        """
        self.nodes.discard(node)
    
    
    def find_ideal_node(self, region=None):
        """
        Finds the least used node in the given region.
        
        Parameters
        ----------
        region : `None`, ``VoiceRegion`` = `None`, Optional
            The region to find a node in. Defaults to `None`.
        
        Returns
        -------
        node : `None`, ``SolarNode``
        """
        nodes = self.available_nodes
        if not nodes:
            return None
        
        if (region is not None):
            region_nodes = [node for node in nodes if node.region is region]
            if region_nodes:
                nodes = region_nodes
        
        return min(nodes, key = node_penalty_key)
    
    
    async def _node_connected(self, node):
        """
        Called when a node is connected from Lavalink.
        
        Parameters
        ----------
        node : `SolarNode`
            The node that has just connected.
        """
        player_queue = self._player_queue
        if (player_queue is not None):
            task_group = TaskGroup(KOKORO, (Task(KOKORO, player.change_node(node)) for player in player_queue))
            self._player_queue = None
            
            failed_task = await task_group.wait_exception()
            if (failed_task is not None):
                task_group.cancel_all()
                failed_task.get_result()
    
    
    async def _node_disconnected(self, node):
        """
        Called when a node is disconnected from Lavalink.
        
        Parameters
        ----------
        node : `SolarNode`
            The node that has just connected.
        """
        players = node.players
        
        if players:
            best_node = self.find_ideal_node(node.region)
            if best_node is None:
                player_queue = self._player_queue
                if player_queue is None:
                    player_queue = list(players)
                    self._player_queue = player_queue
                else:
                    player_queue.extend(players)
            else:
                task_group = TaskGroup(KOKORO, (Task(KOKORO, player.change_node(best_node)) for player in players))
                failed_task = await task_group.wait_exception()
                if (failed_task is not None):
                    task_group.cancel_all()
                    failed_task.get_result()
    
    
    def get_player(self, guild_id):
        """
        Gets a player.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's identifier where the player is.
        
        Returns
        -------
        player : `None`, ``Player``
            Returns `None` if the guild has no player.
        """
        return self.players.get(guild_id, None)
    
    
    async def join_voice(self, channel, *, cls=SolarPlayer):
        """
        Joins a solar player to the channel. If there is an already existing solar player at the respective guild,
        moves it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `tuple` (`int`, `int`)
            The channel to join to.
        cls : ``SolarPlayerBase`` = ``SolarPlayer``, Optional (Keyword only)
            The player's class to create.
        
        Returns
        -------
        solar_player : ``SolarPlayerBase``
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor as `tuple` (`int`, `int`).
        RuntimeError
            - If there are no available nodes.
            - If the ``SolarClient``'s client is already deconstructed.
        """
        if (cls is not SolarPlayer) and (not issubclass(cls, SolarPlayerBase)):
            raise TypeError(
                f'`cls` can be `{SolarPlayerBase.__name__}` subclass, got {cls!r}.'
            )
        
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_guild_connectable)
        
        
        try:
            player = self.players[guild_id]
        except KeyError:
            region = try_get_voice_region(guild_id, channel_id)
            node = self.find_ideal_node(region)
            if node is None:
                raise RuntimeError('No available nodes!')
            
            player, waiter = cls(node, guild_id, channel_id)
            self.players[guild_id] = player
            await waiter
        
        else:
            if player.channel_id != channel_id:
                client = self._client_reference()
                if client is None:
                    raise RuntimeError(f'`{self.__class__.__name__}` client is deconstructed.')
                
                gateway = client.gateway_for(guild_id)
                await gateway.change_voice_state(guild_id, channel_id)
        
        return player
    
    
    def __repr__(self):
        """Returns the solar client's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        node_count = len(self.nodes)
        if node_count:
            repr_parts.append(' node count=')
            repr_parts.append(repr(node_count))
        
        player_count = len(self.players)
        if player_count:
            repr_parts.append(' player count=')
            repr_parts.append(repr(player_count))
        
        player_queue = self._player_queue
        if (player_queue is not None):
            repr_parts.append(' queued up players=')
            repr_parts.append(len(player_queue))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


def node_penalty_key(node):
    """
    Key sued inside of ``SolarClient.find_ideal_node`` when deciding which is the ideal node based on their penalty.
    
    Parameters
    ----------
    node : ``SolarNode``
        A respective node.
    
    Returns
    -------
    penalty : `float`
        The node's penalty.
    """
    return node.penalty



__all__ = ()

APPLICATION_COMMAND_OPTION_OPTIONS_MAX = 25
APPLICATION_COMMAND_OPTION_CHOICES_MAX = 25

APPLICATION_COMMAND_OPTION_MAX_LENGTH_DEFAULT = 0
APPLICATION_COMMAND_OPTION_MAX_LENGTH_MIN = 1
APPLICATION_COMMAND_OPTION_MAX_LENGTH_MAX = 6000

APPLICATION_COMMAND_OPTION_MIN_LENGTH_DEFAULT = 0
APPLICATION_COMMAND_OPTION_MIN_LENGTH_MIN = 0
APPLICATION_COMMAND_OPTION_MIN_LENGTH_MAX = 6000



__all__ = ()

from ...emoji import Emoji, create_emoji_from_exclusive_data, put_exclusive_emoji_data_into
from ...field_parsers import bool_parser_factory, entity_id_parser_factory, force_string_parser_factory
from ...field_putters import bool_optional_putter_factory, entity_id_putter_factory, force_string_putter_factory
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_entity_validator_factory
)

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN

# emoji

def parse_emoji(data):
    """
    Parses out the `emoji` field from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Channel data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    return create_emoji_from_exclusive_data(data)


def put_emoji_into(emoji, data, defaults):
    """
    Puts the `emoji`'s data into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The forum tag's emoji.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (emoji is not None):
        put_exclusive_emoji_data_into(emoji, data)
    
    return data


validate_emoji = nullable_entity_validator_factory('emoji', Emoji)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('forum_tag_id')

# moderated

parse_moderated = bool_parser_factory('moderated', False)
put_moderated_into = bool_optional_putter_factory('moderated', False)
validate_moderated = bool_validator_factory('moderated', False)

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ()

from ...field_parsers import entity_id_parser_factory
from ...field_putters import entity_id_optional_putter_factory
from ...field_validators import bool_validator_factory, entity_id_validator_factory
from ...user import ClientUserBase

from .constants import BOT_ID_KEY, INTEGRATION_ID_KEY, PURCHASABLE_KEY, SUBSCRIPTION_LISTING_ID_KEY

# bot_id

parse_bot_id = entity_id_parser_factory(BOT_ID_KEY)
put_bot_id_into = entity_id_optional_putter_factory(BOT_ID_KEY)
validate_bot_id = entity_id_validator_factory('bot_id', ClientUserBase)

# integration_id

parse_integration_id = entity_id_parser_factory(INTEGRATION_ID_KEY)
put_integration_id_into = entity_id_optional_putter_factory(INTEGRATION_ID_KEY)
validate_integration_id = entity_id_validator_factory('integration_id', NotImplemented, include = 'Integration')

# purchasable

def parse_purchasable(data):
    """
    Parses whether the role is available for purchase.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Role manager data.
    
    Returns
    -------
    purchasable : `bool`
    """
    return PURCHASABLE_KEY in data


def put_purchasable_into(purchasable, data, defaults):
    """
    Puts the given `purchasable` value in the given role manager data.
    
    Parameters
    ----------
    purchasable : `bool`
        Whether the role is available for purchase.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    if purchasable:
        data[PURCHASABLE_KEY] = None
    
    return data

validate_purchasable = bool_validator_factory('purchasable', True)

# subscription_listing_id

parse_subscription_listing_id = entity_id_parser_factory(SUBSCRIPTION_LISTING_ID_KEY)
put_subscription_listing_id_into = entity_id_optional_putter_factory(SUBSCRIPTION_LISTING_ID_KEY)
validate_subscription_listing_id = entity_id_validator_factory('subscription_listing_id')



__all__ = ('merge_update_reaction_mapping',)

def merge_update_reaction_mapping(old_reactions, new_reactions):
    """
    Merges the two reaction mapping, so values wont be overwritten if not required.
    
    Parameters
    ----------
    old_reactions : `None`, ``ReactionMapping``
        The old reactions on a message.
    new_reactions : `None`, ``ReactionMapping``
        The new reactions on a message.
    
    Returns
    -------
    real_reactions : `None`, ``ReactionMapping``
        The real merged reactions on a message.
    """
    if (old_reactions is None):
        return new_reactions
    
    elif (new_reactions is None):
        old_reactions.clear()
        return old_reactions
    
    old_emojis = {*old_reactions.keys()}
    new_emojis = {*new_reactions.keys()}
    
    for emoji in old_emojis - new_emojis:
        del old_reactions[emoji]
    
    for emoji in new_emojis - old_emojis:
        old_reactions[emoji] = new_reactions[emoji].copy()
    
    for emoji in new_emojis & old_emojis:
        old_users = old_reactions[emoji]
        new_users = new_reactions[emoji]
        
        if (len(old_users) != len(new_users)) or (not (old_users >= new_users)):
            old_reactions[emoji] = new_reactions[emoji].copy()
    
    old_reactions._full_check()
    
    return old_reactions



__all__ = ('VoiceChannelEffect',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS
from ...user import create_partial_user_from_id

from ..channel import ChannelType, create_partial_channel_from_id

from .fields import (
    parse_animation_id, parse_animation_type, parse_channel_id, parse_emoji, parse_guild_id, parse_user_id,
    put_animation_id_into, put_animation_type_into, put_channel_id_into, put_emoji_into, put_guild_id_into,
    put_user_id_into, validate_animation_id, validate_animation_type, validate_channel_id, validate_emoji,
    validate_guild_id, validate_user_id
)
from .preinstanced import VoiceChannelEffectAnimationType


class VoiceChannelEffect(EventBase):
    """
    Represents a voice channel effect sent to a voice channel.
    
    Attributes
    ----------
    animation_id : `int`
        The identifier of the animation of the voice channel effect.
    animation_type : ``VoiceChannelEffectAnimationType``
        The animation's type.
    channel_id : `int`
        The channel's identifier where the effect was sent.
    emoji : `None`, ``Emoji``
        The emoji sent.
    guild_id : `int`
        The guild's identifier where the event was sent.
    user_id : `int`
        The user's identifier who sent the effect.
    """
    __slots__ = ('animation_id', 'animation_type', 'channel_id', 'emoji', 'guild_id', 'user_id')
    
    def __new__(
        cls,
        *,
        animation_id = ...,
        animation_type = ...,
        channel_id = ...,
        emoji = ...,
        guild_id = ...,
        user_id = ...,
    ):
        """
        Creates a new voice channel effect event.
        
        Parameters
        ----------
        animation_id : `int`, Optional (Keyword only)
            The identifier of the animation of the voice channel effect.
        animation_type : ``VoiceChannelEffectAnimationType``, `int`, Optional (Keyword only)
            The animation's type.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the effect was sent.
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The emoji sent.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild or its identifier where the event was sent.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or its identifier who sent the effect.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # animation_id
        if animation_id is ...:
            animation_id = 0
        else:
            animation_id = validate_animation_id(animation_id)
        
        # animation_type
        if animation_type is ...:
            animation_type = VoiceChannelEffectAnimationType.premium
        else:
            animation_type = validate_animation_type(animation_type)
        
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.animation_id = animation_id
        self.animation_type = animation_type
        self.channel_id = channel_id
        self.emoji = emoji
        self.guild_id = guild_id
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a voice channel effect event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Voice channel effect event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.animation_id = parse_animation_id(data)
        self.animation_type = parse_animation_type(data)
        self.channel_id = parse_channel_id(data)
        self.emoji = parse_emoji(data)
        self.guild_id = parse_guild_id(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the voice channel effect event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_animation_id_into(self.animation_id, data, defaults)
        put_animation_type_into(self.animation_type, data, defaults)
        put_emoji_into(self.emoji, data, defaults)
        
        if include_internals:
            put_channel_id_into(self.channel_id, data, defaults)
            put_guild_id_into(self.guild_id, data, defaults)
            put_user_id_into(self.user_id, data, defaults)
        
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append(', user_id = ')
        repr_parts.append(repr(self.user_id))
        
        animation_type = self.animation_type
        repr_parts.append(', animation_type = ')
        repr_parts.append(animation_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(animation_type.value))
        
        repr_parts.append(', animation_id = ')
        repr_parts.append(repr(self.animation_id))
        
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 5
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.channel
        yield self.user
        yield self.animation_type
        yield self.animation_type
        yield self.emoji
        
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # animation_id
        if self.animation_id != other.animation_id:
            return False
        
        # animation_type
        if self.animation_type is not other.animation_type:
            return False
        
        # channel_id
        if self.channel_id is not other.channel_id:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # animation_id
        hash_value ^= hash(self.animation_id)
        
        # animation_type
        hash_value ^= hash(self.animation_type)
        
        # channel_id
        hash_value ^= self.channel_id
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the voice channel effect event.
        
        Returns
        -------
        new : `instance<type<new>>`
        """
        new = object.__new__(type(self))
        new.animation_id = self.animation_id
        new.animation_type = self.animation_type
        new.channel_id = self.channel_id
        new.emoji = self.emoji
        new.guild_id = self.guild_id
        new.user_id = self.user_id
        return new
    
    
    def copy_with(
        self,
        *,
        animation_id = ...,
        animation_type = ...,
        channel_id = ...,
        emoji = ...,
        guild_id = ...,
        user_id = ...,
    ):
        """
        Copies new voice channel effect event with the given fields.
        
        Parameters
        ----------
        animation_id : `int`, Optional (Keyword only)
            The identifier of the animation of the voice channel effect.
        animation_type : ``VoiceChannelEffectAnimationType``, `int`, Optional (Keyword only)
            The animation's type.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the effect was sent.
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            The emoji sent.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild or its identifier where the event was sent.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or its identifier who sent the effect.
        
        Returns
        -------
        new : `instance<type<new>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # animation_id
        if animation_id is ...:
            animation_id = self.animation_id
        else:
            animation_id = validate_animation_id(animation_id)
        
        # animation_type
        if animation_type is ...:
            animation_type = self.animation_type
        else:
            animation_type = validate_animation_type(animation_type)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.animation_id = animation_id
        new.animation_type = animation_type
        new.channel_id = channel_id
        new.emoji = emoji
        new.guild_id = guild_id
        new.user_id = user_id
        return new
    
    
    @property
    def channel(self):
        """
        Returns the voice channel effect event's channel.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.guild_voice, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the voice channel effect's guild. If the guild is not cached returns `None`.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the voice channel effect event's user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)
